import os
from tkinter import *
from typing import Optional, Tuple, Union

import cv2
import numpy as np
from PIL import Image, ImageTk
from scipy.spatial.distance import cdist
from shapely.geometry import Polygon

from simba.roi_tools.roi_utils import get_image_from_label
from simba.utils.checks import (check_if_valid_rgb_tuple, check_instance,
                                check_int)
from simba.utils.warnings import ROIWarning

DRAW_FRAME_NAME = "DEFINE SHAPE"

class ROISelectorPolygon(object):

    """
    Class for selecting a polygonal region of interest (ROI) within an image or video frame.
    The selected region vertices are stored in self: polygon_vertices.

    .. video:: tutorials_rst/img/roi/draw_polygon.webm
       :width: 900
       :loop:

    .. seealso::
       For OpenCV based method, see :func:`simba.video_processors.roi_selector_polygon.ROISelectorPolygon`

    :param Toplevel img_window: The Tkinter window where the image is displayed.
    :param int thickness: Thickness of the polygon edges.
    :param int vertice_size: Size of the vertices drawn.
    :param Tuple[int, int, int] clr: RGB color for the polygon.

    :example:
    """

    def __init__(self,
                 img_window: Toplevel,
                 thickness: int = 10,
                 vertice_size: int = 2,
                 clr: Tuple[int, int, int] = (147, 20, 255),
                 tolerance: int = 2) -> None:

        check_instance(source=self.__class__.__name__, instance=img_window, accepted_types=(Toplevel,))
        check_int(name=f'{self.__class__.__name__} thickness', value=thickness, min_value=1)
        check_int(name=f'{self.__class__.__name__} vertice_size', value=vertice_size, min_value=1)
        check_int(name=f'{self.__class__.__name__} tolerance', value=tolerance, min_value=1)
        check_if_valid_rgb_tuple(data=clr)
        self.thickness, self.clr, self.img_window = thickness, clr, img_window
        self.drawing, self.clr, self.thickness, self.vertice_size, self.tolerance = False, clr, thickness, vertice_size, tolerance
        self.polygon_vertices = []

        self.img_lbl = img_window.nametowidget("img_lbl")
        self.img = get_image_from_label(self.img_lbl)
        self.img_cpy = self.img.copy()
        self.w, self.h = self.img.shape[1], self.img.shape[0]

        self.img_window.bind("<ButtonPress-1>", self.on_left_click)
        self.img_window.bind("<B1-Motion>")

    def on_left_click(self, event):
        self.polygon_vertices.append( (event.x, event.y))
        self.img_cpy = cv2.circle(self.img_cpy, center=self.polygon_vertices[-1], radius=self.vertice_size, color=self.clr, thickness=-1, lineType=-1)
        if len(self.polygon_vertices) > 1:
            for i in range(len(self.polygon_vertices) - 1):
                self.img_cpy = cv2.line(self.img_cpy, self.polygon_vertices[i], self.polygon_vertices[i + 1], self.clr, self.thickness)
        self.update_img(self.img_cpy)

    def update_img(self, img):
        img_rgb = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        pil_image = Image.fromarray(img_rgb)
        tk_image = ImageTk.PhotoImage(pil_image)
        self.img_lbl.configure(image=tk_image)
        self.img_lbl.image = tk_image

    def get_attributes(self):
        self.img_window.unbind("<ButtonPress-1>")
        self.img_window.unbind("<B1-Motion>")

        if len(list(set(self.polygon_vertices))) < 3:
            ROIWarning(msg="ROI WARNING: At least 3 unique vertices are needed to form a polygon. Please try again.", source=self.__class__.__name__)
            return False
        else:
            self.polygon = Polygon(np.array(self.polygon_vertices)).simplify(tolerance=self.tolerance, preserve_topology=True)
            self.polygon_vertices = np.array(self.polygon.exterior.coords)
            self.polygon_area = self.polygon.area
            self.polygon_arr = np.array(self.polygon.exterior.coords).astype(np.int32)[1:]
            self.max_vertice_distance = np.max(cdist(self.polygon_vertices, self.polygon_vertices).astype(np.int32))
            try:
                self.polygon_centroid = np.array(self.polygon.centroid).astype(np.int32)
            except TypeError:
                self.polygon_centroid = np.array([self.polygon.centroid.x, self.polygon.centroid.y]).astype(np.int32)
            self.tags = {f'Tag_{cnt}': tuple(y) for cnt, y in enumerate(self.polygon_arr)}
            self.tags['Center_tag'] = tuple(self.polygon_centroid)
            return True

# img = cv2.imread(r"C:\Users\sroni\OneDrive\Desktop\Screenshot 2024-11-15 123805.png")
# root = Toplevel()
# root.title(DRAW_FRAME_NAME)
# img_lbl = Label(root, name='img_lbl')
# img_lbl.pack()
#
# img_rgb = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
# pil_image = Image.fromarray(img_rgb)
# tk_image = ImageTk.PhotoImage(pil_image)
# img_lbl.configure(image=tk_image)
# img_lbl.image = tk_image
#
# _ = ROISelectorPolygon(img_window=root)
# root.mainloop()
#
#     #
    #
    # def get_attributes(self):
    #     if len(list(set(self.polygon_vertices))) < 3:
    #         ROIWarning(msg="ROI WARNING: At least 3 unique vertices are needed to form a polygon. Please try again.", source=self.__class__.__name__)
    #         return False
    #     else:
    #         self.polygon = Polygon(np.array(self.polygon_vertices)).simplify(tolerance=20, preserve_topology=True)
    #         self.polygon_vertices = np.array(self.polygon.exterior.coords)
    #         self.polygon_area = self.polygon.area
    #         self.polygon_arr = np.array(self.polygon.exterior.coords).astype(np.int32)[1:]
    #         self.max_vertice_distance = np.max(cdist(self.polygon_vertices, self.polygon_vertices).astype(np.int32))
    #         self.polygon_centroid = np.array(self.polygon.centroid).astype(int)
    #         self.tags = {f'Tag_{cnt}': tuple(y) for cnt, y in enumerate(self.polygon_arr)}
    #         self.tags['Center_tag'] = tuple(self.polygon_centroid)
    #         return True



__author__ = "Simon Nilsson"



import os
from tkinter import *
from typing import Union

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.data_plotter import DataPlotter
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        SimbaButton, SimbaCheckbox,
                                        SimBADropDown)
from simba.utils.enums import Formats, Keys, Links
from simba.utils.errors import (DuplicationError, InvalidInputError,
                                NoFilesFoundError)
from simba.utils.read_write import get_fn_ext

DECIMALS_OPTIONS = list(range(0, 11))
THICKNESS_OPTIONS = list(range(0, 11))

class DataPlotterPopUp(PopUpMixin, ConfigReader):

    """
    :example:
    >>> _ = DataPlotterPopUp(config_path=r'C:\troubleshooting\RAT_NOR\project_folder\project_config.ini')
    """
    def __init__(self,
                 config_path: Union[str, os.PathLike]):

        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        self.animal_cnt_options = list(range(1, self.animal_cnt + 1))
        if len(self.outlier_corrected_paths) == 0:
            raise NoFilesFoundError(msg=f'Cannot create data plots: No data files found in {self.outlier_corrected_dir} directory.', source=self.__class__.__name__)
        self.file_paths_dict = {get_fn_ext(x)[1]: x for x in self.outlier_corrected_paths}
        self.max_len = max(len(s) for s in list(self.file_paths_dict.keys())) + 5

        PopUpMixin.__init__(self, title="CREATE DATA PLOTS", icon='data_table')
        self.color_lst = list(self.colors_dict.keys())

        self.style_settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="STYLE SETTINGS", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.DATA_TABLES.value, pady=5, padx=5, relief='solid')
        self.resolution_dropdown = SimBADropDown(parent=self.style_settings_frm, dropdown_options=self.resolutions, label='RESOLUTION:', label_width=30, dropdown_width=20, value=self.resolutions[1])
        self.rounding_decimals_dropdown = SimBADropDown(parent=self.style_settings_frm, dropdown_options=DECIMALS_OPTIONS, label='DECIMAL ACCURACY:', label_width=30, dropdown_width=20, value=2)
        self.background_color_dropdown = SimBADropDown(parent=self.style_settings_frm, dropdown_options=self.color_lst, label='BACKGROUND COLOR:', label_width=30, dropdown_width=20, value="White")
        self.header_color_dropdown = SimBADropDown(parent=self.style_settings_frm, dropdown_options=self.color_lst, label='HEADER COLOR:', label_width=30, dropdown_width=20, value="Black")
        self.font_thickness_dropdown = SimBADropDown(parent=self.style_settings_frm, dropdown_options=THICKNESS_OPTIONS, label='FONT THICKNESS: ', label_width=30, dropdown_width=20, value=1)
        self.style_settings_frm.grid(row=0, sticky=NW, pady=10, padx=10)
        self.resolution_dropdown.grid(row=0, sticky=NW)
        self.rounding_decimals_dropdown.grid(row=1, sticky=NW)
        self.background_color_dropdown.grid(row=2, sticky=NW)
        self.header_color_dropdown.grid(row=3, sticky=NW)
        self.font_thickness_dropdown.grid(row=4, sticky=NW)


        self.body_parts_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="CHOOSE BODY-PARTS", icon_name='pose', pady=5, padx=5, relief='solid')
        self.number_of_animals_dropdown = SimBADropDown(parent=self.body_parts_frm, dropdown_options=THICKNESS_OPTIONS, label='# ANIMALS: ', label_width=30, dropdown_width=20, value=1, command=self._create_bp_menu)
        self.body_parts_frm.grid(row=1, column=0, sticky=NW, pady=10, padx=10)
        self.number_of_animals_dropdown.grid(row=0, column=0, sticky=NW)
        self._create_bp_menu(x=1)

        self.settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="VISUALIZATION SETTINGS", icon_name='eye', pady=5, padx=5, relief='solid')
        data_frames_cb, self.data_frames_var = SimbaCheckbox(parent=self.settings_frm, txt="CREATE FRAMES", txt_img='frames', val=False)
        data_videos_cb, self.data_videos_var = SimbaCheckbox(parent=self.settings_frm, txt="CREATE VIDEOS", txt_img='video', val=True)

        self.settings_frm.grid(row=2, column=0, sticky=NW, pady=10, padx=10)
        data_frames_cb.grid(row=0, column=0, sticky=NW)
        data_videos_cb.grid(row=1, column=0, sticky=NW)

        self.run_single_video_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SINGLE VIDEO", icon_name='video', pady=5, padx=5, relief='solid')
        self.run_single_video_btn = SimbaButton(parent=self.run_single_video_frm, img='rocket', txt="CREATE SINGLE VIDEO", font=Formats.FONT_REGULAR.value, cmd=self._run, cmd_kwargs={'multiple': False})
        self.single_video_dropdown = SimBADropDown(parent=self.run_single_video_frm, dropdown_options=list(self.file_paths_dict.keys()), label='VIDEO: ', label_width=30, dropdown_width=self.max_len, value=list(self.file_paths_dict.keys())[0])

        self.run_single_video_frm.grid(row=3, column=0, sticky=NW, pady=10, padx=10)
        self.run_single_video_btn.grid(row=0, column=0, sticky=NW)
        self.single_video_dropdown.grid(row=0, column=1, sticky=NW)

        self.run_multiple_videos = CreateLabelFrameWithIcon(parent=self.main_frm, header="MULTIPLE VIDEOS", icon_name='stack', pady=5, padx=5, relief='solid')
        self.run_multiple_video_btn = SimbaButton(parent=self.run_multiple_videos, img='rocket', txt=f"CREATE MULTIPLE VIDEOS ({len(self.outlier_corrected_paths)} video(s) found)", font=Formats.FONT_REGULAR.value, cmd=self._run, cmd_kwargs={'multiple': True})

        self.run_multiple_videos.grid(row=4, column=0, sticky=NW, pady=10, padx=10)
        self.run_multiple_video_btn.grid(row=0, column=0, sticky=NW)

        #self.main_frm.mainloop()

    def _create_bp_menu(self, x):
        if hasattr(self, "bp_dropdowns"):
            for k, v in self.bp_dropdowns.items():
                self.bp_dropdowns[k].destroy()
                self.bp_colors[k].destroy()
        self.bp_dropdowns, self.bp_colors = {}, {}
        for animal_cnt in range(int(self.number_of_animals_dropdown.getChoices())):
            self.bp_dropdowns[animal_cnt] = SimBADropDown(parent=self.body_parts_frm, dropdown_options=self.body_parts_lst, label=f"BODY-PART {animal_cnt+1}:", label_width=30, dropdown_width=20, value=self.body_parts_lst[animal_cnt])
            self.bp_colors[animal_cnt] = SimBADropDown(parent=self.body_parts_frm, dropdown_options=self.color_lst, label=f"", label_width=2, dropdown_width=20, value=self.color_lst[animal_cnt])
            self.bp_dropdowns[animal_cnt].grid(row=animal_cnt + 1, column=0, sticky=NW)
            self.bp_colors[animal_cnt].grid(row=animal_cnt + 1, column=1, sticky=NW)

    def _run(self, multiple: bool):
        if multiple:
            data_paths = self.outlier_corrected_paths
        else:
            data_paths = [self.file_paths_dict[self.single_video_dropdown.getChoices()]]
        print(data_paths)
        width = int(self.resolution_dropdown.getChoices().split("×")[0])
        height = int(self.resolution_dropdown.getChoices().split("×")[1])
        body_part_attr = []
        for k, v in self.bp_dropdowns.items():
            body_part_attr.append((v.getChoices(), self.colors_dict[self.bp_colors[k].getChoices()]))
        selected_bps = [x[0] for x in body_part_attr]
        if len(list(set(selected_bps))) != len(body_part_attr):
            raise InvalidInputError(msg=f"Please choose unique only body-parts for plotting. Got {len(body_part_attr)} body-parts but only {len(list(set(selected_bps)))} unique", source=DataPlotterPopUp.__class__.__name__)

        bg_color = self.colors_dict[self.background_color_dropdown.getChoices()]
        header_color = self.colors_dict[self.header_color_dropdown.getChoices()]
        font_thickness = int(self.font_thickness_dropdown.getChoices())
        size = (int(width), int(height))
        decimals = int(self.rounding_decimals_dropdown.getChoices())

        if bg_color == header_color:
            raise DuplicationError(msg=f'The header color cannot be the same color as the background color: {bg_color}' , source=self.__class__.__name__)
        for c, i in enumerate(body_part_attr):
            if i[1] == bg_color:
                raise DuplicationError(msg=f'The text color for animal {c+1} cannot be the same color as the background color: {i[1]}', source=self.__class__.__name__)



        if not self.data_videos_var.get() and not self.data_frames_var.get():
            raise InvalidInputError(msg='Both frames and video is set to False, please select one.', source=self.__class__.__name__)

        data_plotter = DataPlotter(config_path=self.config_path,
                                   body_parts=body_part_attr,
                                   data_paths=data_paths,
                                   bg_clr=bg_color,
                                   header_clr=header_color,
                                   font_thickness=font_thickness,
                                   img_size=size,
                                   decimals=decimals,
                                   video_setting=self.data_videos_var.get(),
                                   frame_setting=self.data_frames_var.get())

        _ = data_plotter.run()

#_ = DataPlotterPopUp(config_path=r"C:\troubleshooting\SDS_pre_post\project_folder\project_config.ini")
# _ = DataPlotterPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')



from tkinter import *
from typing import Tuple

import cv2
from PIL import Image, ImageTk

from simba.roi_tools.roi_utils import get_image_from_label
from simba.utils.checks import (check_if_valid_rgb_tuple, check_instance,
                                check_int)
from simba.utils.warnings import ROIWarning

DRAW_FRAME_NAME = "DEFINE SHAPE"

class ROISelector:
    """
    A class to allow users to interactively select a Region of Interest (ROI) rectangle from an image displayed in a Tkinter window.

    .. video:: tutorials_rst/img/roi/draw_rectangle.webm
       :width: 900
       :loop:

    .. seealso::
       For OpenCV based method, see :func:`simba.video_processors.roi_selector.ROISelector`

    :param img_window (Toplevel): The Tkinter Toplevel window containing the image.
    :param thickness (int): The thickness of the ROI selection rectangle.
    :param clr (Tuple[int, int, int]): The color of the ROI selection rectangle in RGB format.

    :example:
    """
    def __init__(self,
                 img_window: Toplevel,
                 thickness: int = 10,
                 clr: Tuple[int, int, int] = (147, 20, 255)):

        check_instance(source=self.__class__.__name__, instance=img_window, accepted_types=(Toplevel,))
        check_int(name=f'{self.__class__.__name__} thickness', value=thickness, min_value=1)
        check_if_valid_rgb_tuple(data=clr)
        self.thickness, self.clr, self.img_window = thickness, clr, img_window

        self.img_lbl = img_window.nametowidget("img_lbl")
        self.img = get_image_from_label(self.img_lbl)
        self.img_cpy = self.img.copy()
        self.w, self.h = self.img.shape[1], self.img.shape[0]

        self.roi_start, self.roi_end = None, None
        self.selecting_roi, self.complete = False, False

        self.img_window.bind("<ButtonPress-1>", self.on_mouse_down)
        self.img_window.bind("<B1-Motion>", self.on_mouse_drag)
        self.img_window.bind("<ButtonRelease-1>", self.on_mouse_up)

    def on_mouse_down(self, event):
        self.roi_start = (event.x, event.y)
        self.roi_end = (event.x, event.y)
        self.selecting_roi = True
#
    def on_mouse_drag(self, event):
        if self.selecting_roi:
            self.roi_end = (event.x, event.y)
            self.img_cpy = self.img.copy()
            cv2.rectangle(self.img_cpy, self.roi_start, self.roi_end, self.clr, self.thickness)
            self.update_image(self.img_cpy)

    def on_mouse_up(self, event):
        self.roi_end = (event.x, event.y)
        self.selecting_roi = False
        self.complete = True
    #
    def update_image(self, img):
        img_rgb = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        pil_image = Image.fromarray(img_rgb)
        tk_image = ImageTk.PhotoImage(pil_image)
        self.img_lbl.configure(image=tk_image)
        self.img_lbl.image = tk_image

    def get_attributes(self):
        self.top_left = min(self.roi_start[0], self.roi_end[0]), min(self.roi_start[1], self.roi_end[1])
        self.bottom_right = max(self.roi_start[0], self.roi_end[0]), max(self.roi_start[1], self.roi_end[1])
        if self.top_left[0] < 0:
            self.top_left = (0, self.top_left[1])
        if self.top_left[1] < 0:
            self.top_left = (self.top_left[0], 0)
        if self.bottom_right[0] < 0:
            self.bottom_right = (0, self.bottom_right[1])
        if self.bottom_right[1] < 0:
            self.bottom_right = (self.bottom_right[0], 0)
        if self.bottom_right[0] > self.w:
            self.bottom_right = (self.w, self.bottom_right[1])
        if self.bottom_right[1] > self.h:
            self.bottom_right = (self.bottom_right[0], self.h)
        self.width = self.bottom_right[0] - self.top_left[0]
        self.height = self.bottom_right[1] - self.top_left[1]
        self.center = (int(self.top_left[0] + (self.width / 2)), int(self.top_left[1] + (self.height / 2)))

        self.bottom_right_tag = (int(self.top_left[0] + self.width), int(self.top_left[1] + self.height))
        self.top_right_tag = (int(self.top_left[0] + self.width), int(self.top_left[1]))
        self.bottom_left_tag = (int(self.top_left[0]), int(self.top_left[1] + self.height))
        self.top_tag = (int(self.top_left[0] + self.width / 2), int(self.top_left[1]))
        self.right_tag = (int(self.top_left[0] + self.width), int(self.top_left[1] + self.height / 2))
        self.left_tag = (int(self.top_left[0]), int(self.top_left[1] + self.height / 2))
        self.bottom_tag = (int(self.top_left[0] + self.width / 2), int(self.top_left[1] + self.height))

        self.img_window.unbind("<ButtonPress-1>")
        self.img_window.unbind("<B1-Motion>")
        self.img_window.unbind("<ButtonRelease-1>")

        if (self.width == 0 and self.height == 0) or (self.width + self.height + self.top_left[0] + self.top_left[1] == 0):
            ROIWarning(msg="ROI WARNING: ROI height and width are both 0. Please try again.", source=self.__class__.__name__)
            return False
        else:
            return True

# img = cv2.imread(r"C:\Users\sroni\OneDrive\Desktop\Screenshot 2024-11-15 123805.png")
# root = Toplevel()
# root.title(DRAW_FRAME_NAME)
# img_lbl = Label(root, name='img_lbl')
# img_lbl.pack()
#
# img_rgb = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
# pil_image = Image.fromarray(img_rgb)
# tk_image = ImageTk.PhotoImage(pil_image)
# img_lbl.configure(image=tk_image)
# img_lbl.image = tk_image
#
# _ = ROISelector(img_window=root)
# root.mainloop()


import os
from typing import Optional, Union

import numpy as np
import pandas as pd

from simba.mixins.geometry_mixin import GeometryMixin
from simba.plotting.geometry_plotter import GeometryPlotter
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_dir_exists, check_int,
                                check_valid_boolean, check_valid_dataframe)
from simba.utils.errors import FrameRangeError
from simba.utils.read_write import (find_core_cnt, get_fn_ext,
                                    get_video_meta_data)

EXPECTED_COLS = ['FRAME', 'CLASS_ID', 'CLASS_NAME', 'CONFIDENCE', 'X1', 'Y1', 'X2', 'Y2', 'X3', 'Y3', 'X4', 'Y4']
FRAME = 'FRAME'
CLASS_ID = 'CLASS_ID'
CONFIDENCE = 'CONFIDENCE'
CLASS_NAME = 'CLASS_NAME'
CORD_FIELDS = ['X1', 'Y1', 'X2', 'Y2', 'X3', 'Y3', 'X4', 'Y4']

class YOLOVisualizer():

    """
    Visualize the results of YOLO bounding box model.

    .. seealso::
       For inference, see :func:`simba.model.yolo_inference.YoloInference`

    .. video:: _static/img/YOLOVisualizer.webm
      :width: 500
      :loop:
      :autoplay:

    :param Union[str, os.PathLike] data_path: Path to YOLO results CSV results. Produced by :func:`simba.bounding_box_tools.yolo.model.inference_yolo`
    :param Union[str, os.PathLike] video_path: Path to the video from which the data was produced.
    :param Union[str, os.PathLike] save_dir: Directory where to save the video results.
    :param Optional[str] palette: Color palette from where to draw the colors for the bounding polygons/boxes. Default: `Set1`.
    :param Optional[int] core_cnt: The number of CPU cores use dto produce the video. -1 for all available cores. Default: -1.
    :param Optional[bool] verbose: If True, prints progress (useful for debugging). Default: False.

    :example:
    >>> test = YOLOVisualizer(data_path=r"/mnt/c/troubleshooting/yolo_inference/08102021_DOT_Rat7_8(2).csv", video_path=r'/mnt/c/troubleshooting/RAT_NOR/project_folder/videos/08102021_DOT_Rat7_8(2).mp4', save_dir="/mnt/c/troubleshooting/yolo_videos")
    >>> test.run()
    """

    def __init__(self,
                 data_path: Union[str, os.PathLike],
                 video_path: Union[str, os.PathLike],
                 save_dir: Union[str, os.PathLike],
                 palette: Optional[str] = 'Set1',
                 core_cnt: Optional[int] = -1,
                 threshold: float = 0.0,
                 thickness: Optional[int] = None,
                 verbose: Optional[bool] = False):



        check_file_exist_and_readable(file_path=data_path)
        self.video_meta_data = get_video_meta_data(video_path=video_path)
        self.data_path, self.video_path = data_path, video_path
        self.video_name = get_fn_ext(filepath=data_path)[1]
        check_int(name=f'{self.__class__.__name__} core_cnt', value=core_cnt, min_value=-1, unaccepted_vals=[0])
        check_float(name=f'{self.__class__.__name__} threshold', value=threshold, min_value=0.0, max_value=1.0)
        self.core_cnt = core_cnt
        if core_cnt == -1 or core_cnt > find_core_cnt()[0]:
            self.core_cnt = find_core_cnt()[0]
        if thickness is not None:
            check_int(name=f'{self.__class__.__name__} thickness', value=thickness, min_value=0, unaccepted_vals=[0])
        check_if_dir_exists(in_dir=save_dir)
        check_valid_boolean(value=[verbose], source=self.__class__.__name__, raise_error=True)
        self.save_dir, self.verbose, self.palette, self.thickness = save_dir, verbose, palette, thickness
        self.threshold = threshold

    def run(self):
        data_df = pd.read_csv(self.data_path, index_col=0)
        check_valid_dataframe(df=data_df, source=self.__class__.__name__, required_fields=EXPECTED_COLS)
        df_frm_cnt = np.unique(data_df[FRAME].values).shape[0]
        if self.video_meta_data['frame_count'] != df_frm_cnt:
            raise FrameRangeError(
                msg=f'The bounding boxes contain data for {df_frm_cnt} frames, while the video is {self.video_meta_data["frame_count"]} frames',
                source=self.__class__.__name__)
        classes = np.unique(data_df[CLASS_NAME].values)
        geometries = []
        for cls in classes:
            cls_df = data_df[data_df[CLASS_NAME] == cls]
            class_id = cls_df[CLASS_ID].iloc[0]
            missing_frms = [x for x in np.arange(0, df_frm_cnt) if x not in cls_df[FRAME].values]
            missing_df = pd.DataFrame(missing_frms, columns=[FRAME])
            missing_df[CLASS_ID], missing_df[CLASS_NAME], missing_df[CONFIDENCE] = class_id, cls, 0
            for cord_col in CORD_FIELDS: missing_df[cord_col] = 0
            cls_df = pd.concat([cls_df, missing_df], axis=0).sort_values(by=[FRAME])
            cls_df.loc[cls_df[CONFIDENCE] < self.threshold, CORD_FIELDS] = -1
            cls_arr = cls_df[CORD_FIELDS].values
            cls_arr = cls_arr.reshape(cls_arr.shape[0], 4, 2)
            geometries.append(GeometryMixin().multiframe_bodyparts_to_polygon(data=cls_arr, video_name=self.video_name, core_cnt=self.core_cnt, verbose=self.verbose))
        plotter = GeometryPlotter(geometries=geometries,
                                  video_name=self.video_path,
                                  core_cnt=self.core_cnt,
                                  save_dir=self.save_dir,
                                  verbose=self.verbose,
                                  colors=[(0, 255, 255)],
                                  thickness=self.thickness,
                                  shape_opacity=0.6)
        plotter.run()









#
# test = YOLOVisualizer(data_path=r"/mnt/d/netholabs/yolo_test/results/2025-04-17_17-05-07.csv",
#                       video_path=r"/mnt/d/netholabs/out/2025-04-17_17-05-07.mp4",
#                       save_dir=r"/mnt/d/netholabs/yolo_test/yolo_videos",
#                       threshold=0.64)
# test.run()


__author__ = "Simon Nilsson"

import os

""" Tkinter pop-up classes for unsupervised ML"""

import threading
from tkinter import *
from typing import Union

import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (DropDownMenu, Entry_Box, FileSelect,
                                        FolderSelect)
from simba.unsupervised.enums import Clustering, UMLOptions, Unsupervised
from simba.unsupervised.hdbscan_clusterer import HDBSCANClusterer
from simba.utils.checks import check_if_dir_exists
from simba.utils.enums import Formats, Options


class FitClusterModelsPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: Union[str, os.PathLike]):
        PopUpMixin.__init__(self, title="CLUSTERING FIT: GRID SEARCH")
        ConfigReader.__init__(self, config_path=config_path)

        self.dataset_frm = LabelFrame(
            self.main_frm,
            text="DATASET",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.data_dir_selected = FolderSelect(
            self.dataset_frm,
            "DATA DIRECTORY (PICKLES): ",
            lblwidth=25,
            initialdir=self.project_path,
        )
        self.save_frm = LabelFrame(
            self.main_frm,
            text="SAVE",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.save_dir = FolderSelect(
            self.save_frm, "SAVE DIRECTORY: ", lblwidth=25, initialdir=self.project_path
        )
        self.algo_frm = LabelFrame(
            self.main_frm,
            text="ALGORITHM",
            pady=5,
            padx=5,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.algo_dropdown = DropDownMenu(
            self.algo_frm,
            "ALGORITHM:",
            UMLOptions.CLUSTERING_ALGO_OPTIONS.value,
            "25",
            com=lambda x: self.show_hyperparameters(),
        )
        self.algo_dropdown.setChoices(UMLOptions.CLUSTERING_ALGO_OPTIONS.value[0])
        self.value_frm = LabelFrame(self.main_frm, fg="black")
        self.value_entry_box = Entry_Box(self.value_frm, "VALUE:", "25")

        self.dataset_frm.grid(row=0, column=0, sticky=NW)
        self.data_dir_selected.grid(row=0, column=0, sticky=NW)
        self.save_frm.grid(row=1, column=0, sticky=NW)
        self.save_dir.grid(row=0, column=0, sticky=NW)
        self.algo_frm.grid(row=2, column=0, sticky=NW)
        self.algo_dropdown.grid(row=0, column=0, sticky=NW)
        self.value_frm.grid(row=3, column=0, sticky=NW)
        self.value_entry_box.grid(row=0, column=0, sticky=NW)
        self.show_hyperparameters()
        self.main_frm.mainloop()

    def show_hyperparameters(self):
        if hasattr(self, "hyperparameters_frm"):
            self.hyperparameters_frm.destroy()
            self.value_frm.destroy()
            self.run_frm.destroy()
        #
        if self.algo_dropdown.getChoices() == Unsupervised.HDBSCAN.value:
            self.hyperparameters_frm = LabelFrame(
                self.main_frm,
                text="GRID SEARCH CLUSTER HYPER-PARAMETERS",
                font=Formats.LABELFRAME_HEADER_FORMAT.value,
                fg="black",
            )

            Label(self.hyperparameters_frm, text=Clustering.ALPHA.value).grid(
                row=1, column=0
            )
            Label(
                self.hyperparameters_frm, text=Clustering.MIN_CLUSTER_SIZE.value
            ).grid(row=1, column=1)
            Label(self.hyperparameters_frm, text=Clustering.MIN_SAMPLES.value).grid(
                row=1, column=2
            )
            Label(self.hyperparameters_frm, text=Clustering.EPSILON.value).grid(
                row=1, column=3
            )

            self.alpha_listbox = Listbox(
                self.hyperparameters_frm, bg="lightgrey", fg="black", height=5, width=15
            )
            self.min_cluster_size_listbox = Listbox(
                self.hyperparameters_frm, bg="lightgrey", fg="black", height=5, width=15
            )
            self.min_samples_listbox = Listbox(
                self.hyperparameters_frm, bg="lightgrey", fg="black", height=5, width=15
            )
            self.epsilon_listbox = Listbox(
                self.hyperparameters_frm, bg="lightgrey", fg="black", height=5, width=15
            )

            alpha_add_btn = Button(
                self.hyperparameters_frm,
                text="ADD",
                fg="blue",
                command=lambda: self.add_to_listbox_from_entrybox(
                    list_box=self.alpha_listbox, entry_box=self.value_entry_box
                ),
            )
            min_cluster_size_add_btn = Button(
                self.hyperparameters_frm,
                text="ADD",
                fg="blue",
                command=lambda: self.add_to_listbox_from_entrybox(
                    list_box=self.min_cluster_size_listbox,
                    entry_box=self.value_entry_box,
                ),
            )
            min_samples_add_btn = Button(
                self.hyperparameters_frm,
                text="ADD",
                fg="blue",
                command=lambda: self.add_to_listbox_from_entrybox(
                    list_box=self.min_samples_listbox, entry_box=self.value_entry_box
                ),
            )
            epsilon_add_btn = Button(
                self.hyperparameters_frm,
                text="ADD",
                fg="blue",
                command=lambda: self.add_to_listbox_from_entrybox(
                    list_box=self.epsilon_listbox, entry_box=self.value_entry_box
                ),
            )

            alpha_remove_btn = Button(
                self.hyperparameters_frm,
                text="REMOVE",
                fg="red",
                command=lambda: self.remove_from_listbox(list_box=self.alpha_listbox),
            )
            min_cluster_size_remove_btn = Button(
                self.hyperparameters_frm,
                text="REMOVE",
                fg="red",
                command=lambda: self.remove_from_listbox(
                    list_box=self.min_cluster_size_listbox
                ),
            )
            min_samples_remove_btn = Button(
                self.hyperparameters_frm,
                text="REMOVE",
                fg="red",
                command=lambda: self.remove_from_listbox(
                    list_box=self.min_samples_listbox
                ),
            )
            epsilon_remove_btn = Button(
                self.hyperparameters_frm,
                text="REMOVE",
                fg="red",
                command=lambda: self.remove_from_listbox(list_box=self.epsilon_listbox),
            )

            self.add_values_to_several_listboxes(
                list_boxes=[
                    self.alpha_listbox,
                    self.min_cluster_size_listbox,
                    self.min_samples_listbox,
                    self.epsilon_listbox,
                ],
                values=[1, 15, 1, 1],
            )

            self.hyperparameters_frm.grid(row=4, column=0, sticky=NW)
            alpha_add_btn.grid(row=2, column=0)
            min_cluster_size_add_btn.grid(row=2, column=1)
            min_samples_add_btn.grid(row=2, column=2)
            epsilon_add_btn.grid(row=2, column=3)

            alpha_remove_btn.grid(row=3, column=0)
            min_cluster_size_remove_btn.grid(row=3, column=1)
            min_samples_remove_btn.grid(row=3, column=2)
            epsilon_remove_btn.grid(row=3, column=3)

            self.alpha_listbox.grid(row=4, column=0, sticky=NW)
            self.min_cluster_size_listbox.grid(row=4, column=1, sticky=NW)
            self.min_samples_listbox.grid(row=4, column=2, sticky=NW)
            self.epsilon_listbox.grid(row=4, column=3, sticky=NW)

            self.create_run_frm(run_function=self.run_hdbscan_clustering)

    def __get_settings(self):
        self.data_directory = self.data_dir_selected.folder_path
        self.save_directory = self.save_dir.folder_path
        check_if_dir_exists(self.data_dir_selected.folder_path)
        check_if_dir_exists(self.save_dir.folder_path)

    def run_hdbscan_clustering(self):
        self.__get_settings()
        alphas = [float(x) for x in self.alpha_listbox.get(0, END)]
        min_cluster_sizes = [int(x) for x in self.min_cluster_size_listbox.get(0, END)]
        min_samples = [int(x) for x in self.min_samples_listbox.get(0, END)]
        epsilons = [float(x) for x in self.epsilon_listbox.get(0, END)]
        hyper_parameters = {
            "alpha": alphas,
            "min_cluster_size": min_cluster_sizes,
            "min_samples": min_samples,
            "cluster_selection_epsilon": epsilons,
        }
        clusterer = HDBSCANClusterer()
        threading.Thread(
            target=clusterer.fit(
                data_path=self.data_directory,
                save_dir=self.save_directory,
                hyper_parameters=hyper_parameters,
            )
        ).start()


# _ = FitClusterModelsPopUp(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/project_config.ini')



import threading
from tkinter import *

from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.ez_path_plot import EzPathPlot
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        Entry_Box, FileSelect)
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_valid_rgb_tuple, check_int, check_str)
from simba.utils.enums import Formats, Keys, Links, Options
from simba.utils.lookups import get_color_dict


class MakePathPlotPopUp(PopUpMixin):
    def __init__(self):
        PopUpMixin.__init__(self, title="CREATE SIMPLE PATH PLOT", size=(500, 300), icon='path_2')
        settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SETTINGS", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.VIDEO_TOOLS.value)
        self.video_path = FileSelect(settings_frm, "VIDEO PATH: ", lblwidth="30", file_types=[("VIDEO FILE", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)])
        self.body_part = Entry_Box(settings_frm, "BODY PART: ", "30")
        self.data_path = FileSelect(settings_frm, "DATA PATH (e.g., H5 or CSV file): ", lblwidth="30")
        color_lst = list(get_color_dict().keys())


        self.background_color = DropDownMenu(settings_frm, "BACKGROUND COLOR: ", color_lst, "30")
        self.background_color.setChoices(choice="White")



        self.line_color = DropDownMenu(settings_frm, "LINE COLOR: ", color_lst, "30")
        self.line_color.setChoices(choice="Red")
        self.line_thickness = DropDownMenu(settings_frm, "LINE THICKNESS: ", list(range(1, 11)), "30")
        self.line_thickness.setChoices(choice=1)
        self.circle_size = DropDownMenu(settings_frm, "CIRCLE SIZE: ", list(range(1, 11)), "30")
        self.last_frm_only_dropdown = DropDownMenu(settings_frm, "LAST FRAME ONLY: ", ["TRUE", "FALSE"], "30")
        self.last_frm_only_dropdown.setChoices("FALSE")
        self.circle_size.setChoices(choice=5)
        settings_frm.grid(row=0, sticky=W)
        self.video_path.grid(row=0, sticky=W)
        self.data_path.grid(row=1, sticky=W)
        self.body_part.grid(row=2, sticky=W)
        self.background_color.grid(row=3, sticky=W)
        self.line_color.grid(row=4, sticky=W)
        self.line_thickness.grid(row=5, sticky=W)
        self.circle_size.grid(row=6, sticky=W)
        self.last_frm_only_dropdown.grid(row=7, sticky=W)
        Label(settings_frm,fg="green",font=Formats.FONT_REGULAR.value,text=" NOTE: For more complex path plots, faster, \n see 'CREATE PATH PLOTS' under the [VISUALIZATIONS] tab after loading your SimBA project").grid(row=8, sticky=W)
        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()

    def run(self):
        data_path = self.data_path.file_path
        video_path = self.video_path.file_path
        background_color = get_color_dict()[self.background_color.getChoices()]
        line_color = get_color_dict()[self.line_color.getChoices()]
        line_thickness = self.line_thickness.getChoices()
        circle_size = self.circle_size.getChoices()
        bp = self.body_part.entry_get
        check_file_exist_and_readable(file_path=data_path)
        check_int(
            name=f"{self.__class__.__name__} line_thickness",
            value=line_thickness,
            min_value=1,
        )
        check_int(
            name=f"{self.__class__.__name__} circle_size",
            value=circle_size,
            min_value=1,
        )
        check_if_valid_rgb_tuple(data=background_color)
        check_if_valid_rgb_tuple(data=line_color)
        check_str(name=f"{self.__class__.__name__} body-part", value=bp)
        last_frm = self.last_frm_only_dropdown.getChoices()
        if last_frm == "TRUE":
            last_frm = True
        else:
            last_frm = False
        plotter = EzPathPlot(
            data_path=data_path,
            video_path=video_path,
            body_part=bp,
            bg_color=background_color,
            line_color=line_color,
            line_thickness=int(line_thickness),
            circle_size=int(circle_size),
            last_frm_only=last_frm,
        )
        threading.Thread(target=plotter.run).start()


#MakePathPlotPopUp()



import os
from tkinter import *
from typing import List, Optional, Union

from simba.data_processors.cue_light_analyzer import CueLightAnalyzer
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import CreateLabelFrameWithIcon, SimBADropDown
from simba.utils.checks import check_if_dir_exists, check_valid_lst
from simba.utils.enums import Links
from simba.utils.errors import NoFilesFoundError, NoROIDataError
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    str_2_bool)


class CueLightDataAnalyzerPopUp(ConfigReader, PopUpMixin):

    """
    :example:
    >>> CueLightDataAnalyzerPopUp(config_path=r"C:\troubleshooting\cue_light\t1\project_folder\project_config.ini", cue_light_names=['cl'])
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 cue_light_names: List[str],
                 data_dir: Optional[Union[str, os.PathLike]] = None):

        ConfigReader.__init__(self, config_path=config_path, read_video_info=True)
        check_valid_lst(data=cue_light_names, source=self.__class__.__name__, valid_dtypes=(str,), min_len=1, raise_error=True)
        self.read_roi_data()
        self.cue_light_names = cue_light_names
        missing_cue_lights = [x for x in cue_light_names if x not in self.roi_names]
        if len(missing_cue_lights) > 0: raise NoROIDataError(msg=f'{len(missing_cue_lights)} cue-lights are not drawn in the SimBA project. The {missing_cue_lights} ROIs cannot be found in the {self.roi_coordinates_path} file.', source=self.__class__.__name__)

        if data_dir is None:
            self.data_dir = self.outlier_corrected_dir
        else:
            check_if_dir_exists(in_dir=data_dir)
            self.data_dir = data_dir

        self.data_paths = find_files_of_filetypes_in_directory(directory=self.data_dir, extensions=[f'.{self.file_type}'], raise_error=False, as_dict=True)
        if len(list(self.data_paths.keys())) == 0 or not os.path.isdir(self.data_dir):
            raise NoFilesFoundError(msg=f'No files detected in directory {self.data_dir}. Perform (or skip outlier correction on imported pose data before performing cue-ligh analysis', source=self.__class__.__name__)

        PopUpMixin.__init__(self, size=(750, 300), title="CUE LIGHT DATA ANALYSIS", icon='data_table')
        self.settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SETTINGS", icon_name='settings', icon_link=Links.CUE_LIGHTS.value)

        self.details_dropdown = SimBADropDown(parent=self.settings_frm, dropdown_options=['TRUE', 'FALSE'], label=f'COMPUTE DETAILED CUE LIGHT BOUT DATA:', label_width=40, dropdown_width=15, value='TRUE')
        self.core_cnt_dropdown = SimBADropDown(parent=self.settings_frm, dropdown_options=list(range(1, self.cpu_cnt+1)), label=f'CPU CORE COUNT:', label_width=40, dropdown_width=15, value=int(self.cpu_cnt/2))
        self.verbose_dropdown = SimBADropDown(parent=self.settings_frm, dropdown_options=['TRUE', 'FALSE'], label=f'VERBOSE:', label_width=40, dropdown_width=15, value='TRUE')

        self.settings_frm.grid(row=0, column=0, sticky=NW)
        self.details_dropdown.grid(row=0, column=0, sticky=NW)
        self.core_cnt_dropdown.grid(row=1, column=0, sticky=NW)
        self.verbose_dropdown.grid(row=2, column=0, sticky=NW)

        self.create_run_frm(run_function=self.run)
        #self.main_frm.mainloop()

    def run(self):
        details = str_2_bool(self.details_dropdown.get_value())
        core_cnt = int(self.core_cnt_dropdown.get_value())
        verbose = str_2_bool(self.verbose_dropdown.get_value())

        cue_light_analyzer = CueLightAnalyzer(config_path=self.config_path,
                                              cue_light_names=self.cue_light_names,
                                              detailed_data=details,
                                              core_cnt=core_cnt,
                                              data_dir=self.outlier_corrected_dir,
                                              verbose=verbose)
        cue_light_analyzer.run()


__author__ = "Simon Nilsson"


import functools
import multiprocessing
import os
from typing import Dict, Optional, Union

import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.utils.checks import check_float, check_if_dir_exists, check_int
from simba.utils.enums import ConfigKey, Defaults, Dtypes
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_core_cnt,
                                    find_files_of_filetypes_in_directory,
                                    get_fn_ext, read_config_entry, read_df,
                                    write_df)


def _location_outlier_corrector(data_path: str,
                                config: FeatureExtractionMixin,
                                animal_bp_dict: dict,
                                outlier_dict: dict,
                                file_type: str,
                                save_dir: str,
                                criterion: float):

    def __find_location_outliers(bp_dict: dict, animal_criteria: dict):
        above_criteria_dict, below_criteria_dict = {}, {}
        for animal_name, animal_data in bp_dict.items():
            animal_criterion = animal_criteria[animal_name]
            above_criteria_dict[animal_name]= {}
            for first_bp_cnt, (first_body_part_name, first_bp_cords) in enumerate(animal_data.items()):
                second_bp_names = [x for x in list(animal_data.keys()) if x != first_body_part_name]
                above_criterion_frms = []
                for second_bp_cnt, second_bp in enumerate(second_bp_names):
                    second_bp_cords = animal_data[second_bp]
                    distances = config.framewise_euclidean_distance(location_1=first_bp_cords, location_2=second_bp_cords, px_per_mm=1.0, centimeter=False)
                    above_criterion_frms.extend(np.argwhere(distances > animal_criterion).flatten())
                unique, counts = np.unique(above_criterion_frms, return_counts=True)
                above_criteria_dict[animal_name][first_body_part_name] = np.sort(unique[counts > 1])
        return above_criteria_dict

    def __correct_outliers(df: pd.DataFrame, above_criteria_dict: dict):
        for animal_name, animal_data in above_criteria_dict.items():
            for body_part_name, frm_idx in animal_data.items():
                col_names = [f'{body_part_name}_x', f'{body_part_name}_y']
                if len(frm_idx) > 0:
                    df.loc[frm_idx, col_names] = np.nan
        return df.fillna(method='ffill', axis=0).fillna(0)

    video_timer = SimbaTimer(start=True)
    _, video_name, _ = get_fn_ext(data_path)
    print(f"Processing video {video_name}..")
    save_path = os.path.join(save_dir, f"{video_name}.{file_type}")
    above_criterion_dict, below_criterion_dict, animal_criteria, bp_dict = {}, {}, {}, {}
    df = read_df(data_path, file_type)
    for animal_name, animal_bps in outlier_dict.items():
        animal_bp_distances = np.sqrt((df[animal_bps["bp_1"] + "_x"] - df[animal_bps["bp_2"] + "_x"]) ** 2 + (df[animal_bps["bp_1"] + "_y"] - df[animal_bps["bp_2"] + "_y"]) ** 2)
        animal_criteria[animal_name] = (animal_bp_distances.mean() * criterion)
    for animal_name, animal_bps in animal_bp_dict.items():
        bp_col_names = np.array([[i, j] for i, j in zip(animal_bps["X_bps"], animal_bps["Y_bps"])]).ravel()
        animal_arr = df[bp_col_names].to_numpy()
        bp_dict[animal_name] = {}
        for bp_cnt, bp_col_start in enumerate(range(0, animal_arr.shape[1], 2)):
            bp_name = animal_bps["X_bps"][bp_cnt][:-2]
            bp_dict[animal_name][bp_name] = animal_arr[:, bp_col_start: bp_col_start + 2]

    above_criteria_dict = __find_location_outliers(bp_dict=bp_dict, animal_criteria=animal_criteria)
    df = __correct_outliers(df=df, above_criteria_dict=above_criteria_dict)
    write_df(df=df, file_type=file_type, save_path=save_path)
    video_timer.stop_timer()
    print(f"Corrected location outliers for file {video_name} (elapsed time: {video_timer.elapsed_time_str}s)...")
    return video_name, above_criteria_dict, len(df)

class OutlierCorrecterLocationMultiprocess(ConfigReader, FeatureExtractionMixin):
    """
    Detect and amend outliers in pose-estimation data based in the location of the body-parts
    in the current frame relative to the location of the body-part in the preceding frame using heuristic rules.

    Uses heuristic rules critera is grabbed from the SimBA project project_config.ini under the [Outlier settings] header.

    .. note::
       `Documentation <https://github.com/sgoldenlab/simba/blob/master/misc/Outlier_settings.pdf>`_.

    .. image:: _static/img/location_outlier.png
       :width: 500
       :align: center

    :param Union[str, os.PathLike] config_path: path to SimBA project config file in Configparser format
    :param Optional[Union[str, os.PathLike]] data_dir: The directory storing the input data. If None, then the ``outlier_corrected_movement`` directory of the SimBA project.
    :param Optional[Union[str, os.PathLike]] save_dir: The directory to store the results. If None, then the ``outlier_corrected_movement_location`` directory of the SimBA project.
    :param Optional[int] core_cnt: The number of cores to use. If -1, then all available cores. Default: -1.
    :param Optional[Dict[str, Dict[str, str]]] animal_dict: Dictionary holding the animal names, and the two body-parts to use to measure the mean or median size of the animals. If None, grabs the info from the SimBA project config.
    :param Optional[float] criterion: The criterion multiplier. If None, grabs the info from the SimBA project config.

    :example:
    >>> _ = OutlierCorrecterLocationMultiprocess(config_path='MyProjectConfig').run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 data_dir: Optional[Union[str, os.PathLike]] = None,
                 save_dir: Optional[Union[str, os.PathLike]] = None,
                 core_cnt: Optional[int] = -1,
                 animal_dict: Optional[Dict[str, Dict[str, str]]] = None,
                 criterion: Optional[float] = None):

        ConfigReader.__init__(self, config_path=config_path, create_logger=False, read_video_info=False)
        FeatureExtractionMixin.__init__(self)
        if not os.path.exists(self.outlier_corrected_dir):
            os.makedirs(self.outlier_corrected_dir)
        if criterion is None:
            self.criterion = read_config_entry(self.config, ConfigKey.OUTLIER_SETTINGS.value, ConfigKey.LOCATION_CRITERION.value, Dtypes.FLOAT.value)
        else:
            check_float(name=f'{criterion} criterion', value=criterion, min_value=10e-10)
            self.criterion = criterion
        if data_dir is not None:
            check_if_dir_exists(in_dir=data_dir, source=self.__class__.__name__)
            self.data_dir = data_dir
        else:
            self.data_dir = self.outlier_corrected_movement_dir
        if save_dir is not None:
            check_if_dir_exists(in_dir=save_dir, source=self.__class__.__name__)
            self.save_dir = save_dir
        else:
            self.save_dir = self.outlier_corrected_dir
        check_int(name=f'{self.__class__.__name__} core_cnt', value=core_cnt, min_value=-1, unaccepted_vals=[0])
        self.core_cnt = core_cnt
        if self.core_cnt == -1:
            self.core_cnt = find_core_cnt()[0]

        self.above_criterion_dict_dict, self.below_criterion_dict_dict = {},{}
        if animal_dict is None:
            self.outlier_bp_dict = {}
            if self.animal_cnt == 1:
                self.animal_id = read_config_entry(self.config, ConfigKey.MULTI_ANIMAL_ID_SETTING.value, ConfigKey.MULTI_ANIMAL_IDS.value, Dtypes.STR.value)
                if self.animal_id != "None":
                    self.animal_bp_dict[self.animal_id] = self.animal_bp_dict.pop("Animal_1")

            for animal_name in self.animal_bp_dict.keys():
                self.outlier_bp_dict[animal_name] = {}
                self.outlier_bp_dict[animal_name]["bp_1"] = read_config_entry(self.config, ConfigKey.OUTLIER_SETTINGS.value, "location_bodypart1_{}".format(animal_name.lower()),"str")
                self.outlier_bp_dict[animal_name]["bp_2"] = read_config_entry(self.config, ConfigKey.OUTLIER_SETTINGS.value, "location_bodypart2_{}".format(animal_name.lower()),"str")
        else:
            self.outlier_bp_dict = animal_dict

    def run(self):
        self.logs, self.frm_cnts = {}, {}
        data_paths = find_files_of_filetypes_in_directory(directory=self.data_dir, extensions=[f'.{self.file_type}'], raise_error=True)
        data_path_tuples = [(x) for x in data_paths]
        with multiprocessing.Pool(self.core_cnt, maxtasksperchild=Defaults.MAXIMUM_MAX_TASK_PER_CHILD.value) as pool:
            constants = functools.partial(_location_outlier_corrector,
                                          config=self,
                                          animal_bp_dict=self.animal_bp_dict,
                                          outlier_dict=self.outlier_bp_dict,
                                          save_dir=self.save_dir,
                                          file_type=self.file_type,
                                          criterion=self.criterion)
            for cnt, (video_name, above_critera_dict, frm_cnt) in enumerate(pool.imap(constants, data_path_tuples, chunksize=1)):
                self.frm_cnts[video_name] = frm_cnt
                self.logs[video_name] = above_critera_dict
                print(f"Video {video_name} complete...")
        self.__save_log_file()

    def __save_log_file(self):
        out_df = pd.DataFrame(columns=['VIDEO', 'ANIMAL', 'BODY-PART', 'CORRECTION COUNT', 'CORRECTION RATIO'])
        for video_name, video_data in self.logs.items():
            for animal_name, animal_data in video_data.items():
                for bp_name, bp_data in animal_data.items():
                    correction_ratio = round(len(bp_data) / self.frm_cnts[video_name], 6)
                    out_df.loc[len(out_df)] = [video_name, animal_name, bp_name, len(bp_data), correction_ratio]
        self.logs_path = os.path.join(self.logs_path, f"Outliers_location_{self.datetime}.csv")
        out_df.to_csv(self.logs_path)
        self.timer.stop_timer()
        stdout_success(msg='Log for corrected "location outliers" saved in project_folder/logs', elapsed_time=self.timer.elapsed_time_str)

# if __name__ == "__main__":
#     test = OutlierCorrecterLocationMultiprocess(config_path=r"C:\troubleshooting\two_black_animals_14bp\project_folder\project_config.ini")
#     #test = OutlierCorrecterLocationMultiprocess(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini")
#     test.run()


# test = OutlierCorrecterLocation(config_path='/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/project_config.ini')
# test.run()

# test = OutlierCorrecterLocation(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# test.correct_location_outliers()



__author__ = "Simon Nilsson"

import time
import warnings

warnings.simplefilter(action="ignore", category=FutureWarning)
import functools
import multiprocessing
import os
import platform
from typing import List, Optional, Union

import cv2
import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_file_exist_and_readable, check_int, check_str,
    check_that_column_exist, check_valid_lst)
from simba.utils.data import create_color_palette, detect_bouts
from simba.utils.enums import Formats, Options
from simba.utils.errors import NoSpecifiedOutputError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder,
                                    create_directory, find_core_cnt,
                                    get_fn_ext, read_df)

HEIGHT = "height"
WIDTH = "width"
FONT_ROTATION = "font rotation"
FONT_SIZE = "font size"
STYLE_KEYS = [HEIGHT, WIDTH, FONT_ROTATION, FONT_SIZE]

def gantt_creator_mp(data: np.array,
                     frame_setting: bool,
                     video_setting: bool,
                     video_save_dir: str,
                     frame_folder_dir: str,
                     bouts_df: pd.DataFrame,
                     clf_names: list,
                     fps: int,
                     video_name: str,
                     width: int,
                     height: int,
                     font_size: int,
                     font_rotation: int,
                     palette: np.ndarray):

    batch_id, frame_rng = data[0], data[1]
    start_frm, end_frm, current_frm = frame_rng[0], frame_rng[-1], frame_rng[0]
    video_writer = None
    if video_setting:
        fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        video_save_path = os.path.join(video_save_dir, f"{batch_id}.mp4")
        video_writer = cv2.VideoWriter(video_save_path, fourcc, fps, (width, height))

    while current_frm <= end_frm:
        bout_rows = bouts_df.loc[bouts_df["End_frame"] <= current_frm]
        plot = PlottingMixin().make_gantt_plot(x_length=current_frm + 1,
                                               bouts_df=bout_rows,
                                               clf_names=clf_names,
                                               fps=fps,
                                               width=width,
                                               height=height,
                                               font_size=font_size,
                                               font_rotation=font_rotation,
                                               video_name=video_name,
                                               save_path=None,
                                               palette=palette)
        current_frm += 1
        if frame_setting:
            frame_save_path = os.path.join(frame_folder_dir, f"{current_frm}.png")
            cv2.imwrite(frame_save_path, plot)
        if video_setting:
            video_writer.write(plot)
        print(f"Gantt frame created: {current_frm + 1}, Video: {video_name}, Processing core: {batch_id + 1}")

    if video_setting:
        video_writer.release()

    return batch_id


class GanttCreatorMultiprocess(ConfigReader, PlottingMixin):
    """
    Multiprocess creation of classifier gantt charts in video and/or image format.
    See meth:`simba.gantt_creator.GanttCreatorSingleProcess` for single-process class.

    ..note::
       `GitHub gantt tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#gantt-plot>`__.

    .. image:: _static/img/gantt_plot.png
       :width: 300
       :align: center

    :param str config_path: path to SimBA project config file in Configparser format.
    :param bool frame_setting: If True, creates individual frames.
    :param bool last_frm_setting: If True, creates single .png image representing entire video.
    :param bool video_setting: If True, creates videos
    :param dict style_attr: Attributes of gannt chart (size, font size, font rotation etc).
    :param List[str] files_found: File paths representing files with machine predictions e.g., ['project_folder/csv/machine_results/My_results.csv']
    :param int cores: Number of cores to use.

    :examples:
    >>> gantt_creator = GanttCreatorMultiprocess(config_path='project_folder/project_config.ini', frame_setting=False, video_setting=True, files_found=['project_folder/csv/machine_results/Together_1.csv'], cores=5, style_attr={'width': 640, 'height': 480, 'font size': 8, 'font rotation': 45}).run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 data_paths: List[Union[str, os.PathLike]],
                 frame_setting: Optional[bool] = False,
                 video_setting: Optional[bool] = False,
                 last_frm_setting: Optional[bool] = True,
                 width: int = 640,
                 height: int = 480,
                 font_size: int = 8,
                 font_rotation: int = 45,
                 palette: str = 'Set1',
                 core_cnt: Optional[int] = -1):

        check_file_exist_and_readable(file_path=config_path)
        if (not frame_setting) and (not video_setting) and (not last_frm_setting):
            raise NoSpecifiedOutputError(msg="SIMBA ERROR: Please select gantt videos, frames, and/or last frame.", source=self.__class__.__name__)
        check_file_exist_and_readable(file_path=config_path)
        check_valid_lst(data=data_paths, source=f'{self.__class__.__name__} data_paths', valid_dtypes=(str,), min_len=1)
        check_int(value=width, min_value=1, name=f'{self.__class__.__name__} width')
        check_int(value=height, min_value=1, name=f'{self.__class__.__name__} height')
        check_int(value=font_size, min_value=1, name=f'{self.__class__.__name__} font_size')
        check_int(value=font_rotation, min_value=0, max_value=180, name=f'{self.__class__.__name__} font_rotation')
        palettes = Options.PALETTE_OPTIONS_CATEGORICAL.value + Options.PALETTE_OPTIONS.value
        check_str(name=f'{self.__class__.__name__} palette', value=palette, options=palettes)
        for file_path in data_paths: check_file_exist_and_readable(file_path=file_path)
        check_int(name=f"{self.__class__.__name__} core_cnt",value=core_cnt, min_value=-1, unaccepted_vals=[0], max_value=find_core_cnt()[0])
        self.core_cnt = find_core_cnt()[0] if core_cnt == -1 or core_cnt > find_core_cnt()[0] else core_cnt
        self.width, self.height, self.font_size, self.font_rotation = width, height, font_size, font_rotation
        ConfigReader.__init__(self, config_path=config_path, create_logger=False)
        PlottingMixin.__init__(self)
        self.clr_lst = create_color_palette(pallete_name=palette, increments=len(self.body_parts_lst) + 1, as_int=True, as_rgb_ratio=True)
        self.frame_setting, self.video_setting, self.data_paths, self.last_frm_setting = frame_setting, video_setting,data_paths, last_frm_setting
        if not os.path.exists(self.gantt_plot_dir): os.makedirs(self.gantt_plot_dir)
        if platform.system() == "Darwin":
            multiprocessing.set_start_method("spawn", force=True)
        print(f"Processing {len(self.data_paths)} video(s)...")

    def run(self):
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.data_paths)
        for file_cnt, file_path in enumerate(self.data_paths):
            video_timer = SimbaTimer(start=True)
            _, self.video_name, _ = get_fn_ext(file_path)
            self.data_df = read_df(file_path, self.file_type)
            check_that_column_exist(df=self.data_df, column_name=self.clf_names, file_name=file_path)
            print(f"Processing video {self.video_name}, Frame count: {len(self.data_df)} (Video {(file_cnt + 1)}/{len(self.data_paths)})...")
            self.video_info_settings, _, self.fps = self.read_video_info(video_name=self.video_name)
            self.bouts_df = detect_bouts(data_df=self.data_df, target_lst=list(self.clf_names), fps=int(self.fps))
            self.temp_folder = os.path.join(self.gantt_plot_dir, self.video_name, "temp")
            self.save_frame_folder_dir = os.path.join(self.gantt_plot_dir, self.video_name)
            if self.frame_setting:
                create_directory(paths=self.save_frame_folder_dir, overwrite=True)
            if self.video_setting:
                create_directory(paths=self.temp_folder)
                self.save_video_path = os.path.join(self.gantt_plot_dir, f"{self.video_name}.mp4")

            if self.last_frm_setting:
                self.make_gantt_plot(x_length=len(self.data_df),
                                     bouts_df=self.bouts_df,
                                     clf_names=self.clf_names,
                                     fps=self.fps,
                                     width=self.width,
                                     height=self.height,
                                     font_size=self.font_size,
                                     font_rotation=self.font_rotation,
                                     video_name=self.video_name,
                                     save_path=os.path.join(self.gantt_plot_dir, f"{self.video_name}_final_image.png"),
                                     palette=self.clr_lst)

            if self.video_setting or self.frame_setting:
                frame_data = np.array_split(list(range(0, len(self.data_df))), self.core_cnt)
                frame_data = [(i, x) for i, x in enumerate(frame_data)]
                print(f"Creating gantt, multiprocessing (chunksize: {(self.multiprocess_chunksize)}, cores: {self.core_cnt})...")
                with multiprocessing.Pool(self.core_cnt, maxtasksperchild=self.maxtasksperchild) as pool:
                    constants = functools.partial(gantt_creator_mp,
                                                  video_setting=self.video_setting,
                                                  frame_setting=self.frame_setting,
                                                  video_save_dir=self.temp_folder,
                                                  frame_folder_dir=self.save_frame_folder_dir,
                                                  bouts_df=self.bouts_df,
                                                  clf_names=self.clf_names,
                                                  fps=self.fps,
                                                  width=self.width,
                                                  height=self.height,
                                                  font_size=self.font_size,
                                                  font_rotation=self.font_rotation,
                                                  video_name=self.video_name,
                                                  palette=self.clr_lst)
                    for cnt, result in enumerate(pool.imap(constants, frame_data, chunksize=self.multiprocess_chunksize)):
                        print(f'Batch {result+1/self.core_cnt} complete...')
                pool.terminate()
                pool.join()
                if self.video_setting:
                    print(f"Joining {self.video_name} multiprocessed video...")
                    concatenate_videos_in_folder(in_folder=self.temp_folder, save_path=self.save_video_path)
                video_timer.stop_timer()
                print(f"Gantt video {self.video_name} complete (elapsed time: {video_timer.elapsed_time_str}s) ...")

        self.timer.stop_timer()
        stdout_success(msg=f"Gantt visualizations for {len(self.data_paths)} videos created in {self.gantt_plot_dir} directory", elapsed_time=self.timer.elapsed_time_str)


# if __name__ == "__main__":
#     test = GanttCreatorMultiprocess(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
#                                     frame_setting=False,
#                                     video_setting=False,
#                                     data_paths=[r"C:\troubleshooting\mitra\project_folder\csv\machine_results\592_MA147_Gq_CNO_0515.csv"],
#                                     last_frm_setting=True,
#                                     width=640,
#                                     height= 480,
#                                     font_size=10,
#                                     font_rotation= 45)
#     test.run()




# test = GanttCreatorMultiprocess(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/RAT_NOR/project_folder/project_config.ini',
#                                 frame_setting=False,
#                                 video_setting=True,
#                                 data_paths=['/Users/simon/Desktop/envs/simba/troubleshooting/RAT_NOR/project_folder/csv/machine_results/2022-06-20_NOB_DOT_4.csv'],
#                                 cores=5,
#                                 last_frm_setting=False,
#                                 style_attr={'width': 640, 'height': 480, 'font size': 10, 'font rotation': 65})
# test.run()


# test = GanttCreatorMultiprocess(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                 frame_setting=False,
#                                 video_setting=True,
#                                 data_paths=['/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'],
#                                 last_frm_setting=True,
#                                 style_attr={'width': 640, 'height': 480, 'font size': 10, 'font rotation': 65},
#                                 cores=2)
# test.run()



import itertools
import json
import os
from datetime import datetime
from typing import Optional, Union

import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import cv2

from simba.mixins.image_mixin import ImageMixin
from simba.third_party_label_appenders.transform.utils import b64_to_arr
from simba.utils.checks import (check_if_dir_exists,
                                check_if_keys_exist_in_dict, check_str,
                                check_valid_boolean)
from simba.utils.enums import Options
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    img_array_to_clahe, read_json)
from simba.utils.warnings import NoDataFoundWarning


class Labelme2DLC:

    """
    Convert labels from labelme format to DLC annotation format.

    .. note::
        See `labelme GitHub repo <https://github.com/wkentaro/labelme>`__.

    .. seealso::
       For DLC -> Labelme annotation conversion, see :func:`simba.third_party_label_appenders.transform.dlc_to_labelme.DLC2Labelme`

    :param Union[str, os.PathLike] labelme_dir: Directory with labelme json files.
    :param Optional[str] scorer: Name of the scorer (anticipated by DLC as header)
    :param bool greyscale: If True, convert images to grayscale.
    :param bool clahe: If True, apply CLAHE (Contrast Limited Adaptive Histogram Equalization).
    :param bool verbose: If True, prints progress.
    :param Optional[Union[str, os.PathLike]] save_dir: Directory where to save the DLC annotations. If None, then same directory as labelme_dir with `_dlc_annotations` suffix.
    :return: None


    :example:
    >>> labelme_dir = r"D:\platea\ts_annotations"
    >>> runner = Labelme2DLC(labelme_dir=labelme_dir)
    >>> runner.run()
    """

    def __init__(self,
                 labelme_dir: Union[str, os.PathLike],
                 scorer: str = 'SN',
                 greyscale: bool = False,
                 clahe: bool = False,
                 verbose: bool = True,
                 save_dir: Optional[Union[str, os.PathLike]] = None) -> None:

        check_if_dir_exists(in_dir=labelme_dir)
        check_str(name=f'{self.__class__.__name__} scorer', value=scorer)
        self.annotation_paths = find_files_of_filetypes_in_directory(directory=labelme_dir, extensions=['.json'], raise_error=True, as_dict=True)
        if save_dir is None:
            self.save_dir = os.path.join(os.path.dirname(labelme_dir), os.path.basename(labelme_dir) + f'_dlc_annotations_{datetime.now().strftime("%Y%m%d%H%M%S")}')
            if not os.path.isdir(self.save_dir): os.makedirs(self.save_dir)
        else:
            check_if_dir_exists(in_dir=save_dir)
            self.save_dir = save_dir
        check_valid_boolean(value=greyscale, source=f'{self.__class__.__name__} greyscale', raise_error=True)
        check_valid_boolean(value=clahe, source=f'{self.__class__.__name__} clahe', raise_error=True)
        check_valid_boolean(value=verbose, source=f'{self.__class__.__name__} verbose', raise_error=True)
        self.clahe, self.greyscale, self.scorer, self.labelme_dir, self.verbose = clahe, greyscale, scorer, labelme_dir, verbose
        self.file_cnt = len(list(self.annotation_paths.keys()))


    def run(self):
        timer = SimbaTimer(start=True)
        results_dict, images = {}, {}
        for file_cnt, (file_name, annot_path) in enumerate(self.annotation_paths.items()):
            if self.verbose:
                print(f'Reading labelme file {file_cnt+1}/{self.file_cnt}...')
            annot_data = read_json(x=annot_path)
            check_if_keys_exist_in_dict(data=annot_data, key=['shapes', 'imageData', 'imagePath'], name=annot_path)
            img_name = os.path.basename(annot_data['imagePath'])
            images[img_name] = b64_to_arr(annot_data['imageData'])
            if self.greyscale:
                images[img_name] = ImageMixin.img_to_greyscale(img=images[img_name])
            if self.clahe:
                images[img_name] = img_array_to_clahe(img=images[img_name])
            for bp_data in annot_data['shapes']:
                check_if_keys_exist_in_dict(data=bp_data, key=['label', 'points'], name=annot_path)
                point_x, point_y = bp_data['points'][0][0], bp_data['points'][0][1]
                lbl = bp_data['label']
                id = os.path.join('labeled-data', os.path.basename(self.labelme_dir), img_name)
                if id not in results_dict.keys():
                    results_dict[id] = {f'{lbl}': {'x': point_x, 'y': point_y}}
                else:
                    results_dict[id].update({f'{lbl}': {'x': point_x, 'y': point_y}})

        bp_names = set()
        for img, bp in results_dict.items(): bp_names.update(set(bp.keys()))
        col_names = list(itertools.product(*[[self.scorer], bp_names, ['x', 'y']]))
        columns = pd.MultiIndex.from_tuples(col_names)
        results = pd.DataFrame(columns=columns)
        results.columns.names = ['scorer', 'bodyparts', 'coords']
        for img, bp_data in results_dict.items():
            for bp_name, bp_cords in bp_data.items():
                results.at[img, (self.scorer, bp_name, 'x')] = bp_cords['x']
                results.at[img, (self.scorer, bp_name, 'y')] = bp_cords['y']

        for img_cnt, (img_name, img) in enumerate(images.items()):
            if self.verbose:
                print(f'Saving DLC file {img_cnt+1}/{self.file_cnt}...')
            img_save_path = os.path.join(self.save_dir, img_name)
            cv2.imwrite(img_save_path, img)
        save_path = os.path.join(self.save_dir, f'CollectedData_{self.scorer}.csv')
        results.to_csv(save_path)
        timer.stop_timer()
        if self.verbose:
            stdout_success(msg=f'DLC annotations for {self.file_cnt} images saved in directory {self.save_dir}', elapsed_time=timer.elapsed_time_str, source=self.__class__.__name__)


# labelme_dir = r"D:\platea\ts_annotations"
# runner = Labelme2DLC(labelme_dir=labelme_dir)
# runner.run()


import os
from tkinter import filedialog
from typing import Union

from simba.labelling.standard_labeller import LabellingInterface
from simba.utils.checks import (check_file_exist_and_readable,
                                check_valid_boolean)
from simba.utils.enums import ConfigKey, Dtypes, Options
from simba.utils.errors import NoFilesFoundError
from simba.utils.read_write import (get_video_meta_data, read_config_entry,
                                    read_config_file)


class SelectLabellingVideoPupUp():

    """
    Launch PopUp to select video for labelling

    :example:
    >>> SelectLabellingVideoPupUp(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini", continuing=False)
    """
    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 continuing: bool = False):

        check_file_exist_and_readable(file_path=config_path)
        check_valid_boolean(value=[continuing], source=f'{self.__class__.__name__} config_path', raise_error=True)
        if not continuing: win_title = "SELECT VIDEO FILE - NEW VIDEO ANNOTATION"
        else: win_title = "SELECT VIDEO FILE - CONTINUING VIDEO ANNOTATION"
        config = read_config_file(config_path=config_path)
        project_path = read_config_entry(config, ConfigKey.GENERAL_SETTINGS.value, ConfigKey.PROJECT_PATH.value, data_type=Dtypes.STR.value)
        file_type = read_config_entry(config, ConfigKey.GENERAL_SETTINGS.value, ConfigKey.FILE_TYPE.value, data_type=Dtypes.STR.value)
        video_dir = os.path.join(project_path, 'videos')
        targets_inserted_dir = os.path.join(project_path, 'csv', 'targets_inserted')
        video_dir = None if not os.path.isdir(video_dir) else video_dir
        video_file_path = filedialog.askopenfilename(filetypes=[("VIDEO FILE", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)], initialdir=video_dir, title=win_title)
        check_file_exist_and_readable(file_path=video_file_path)
        video_meta = get_video_meta_data(video_file_path)
        print(f"INITIATING ANNOTATION INTERFACE FOR {video_meta['video_name']} \n VIDEO INFO: {video_meta}")

        if continuing:
            self.targets_inserted_file_path = os.path.join(targets_inserted_dir, f"{video_meta['video_name']}.{file_type}")
            if not os.path.isfile(self.targets_inserted_file_path):
                raise NoFilesFoundError( msg=f'When continuing annotations, SimBA expects a file at {self.targets_inserted_file_path}. SimBA could not find this file.', source=self.__class__.__name__)

        _ = LabellingInterface(config_path=config_path,
                               file_path=video_file_path,
                               thresholds=None,
                               continuing=continuing)



#SelectLabellingVideoPupUp(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini", continuing=False)


import os
from itertools import combinations, product
from typing import List, Union

import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

from simba.mixins.config_reader import ConfigReader
from simba.mixins.statistics_mixin import Statistics
from simba.unsupervised.enums import Clustering, Unsupervised
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists,
                                check_if_keys_exist_in_dict, check_valid_lst)
from simba.utils.enums import Formats
from simba.utils.errors import CountError
from simba.utils.printing import stdout_success
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    get_fn_ext, read_pickle)

ADJ_MUTUAL_INFO = "adjusted mutual information"
FOWLKES_MALLOWS = "fowlkes mallows"
ADJ_RAND_INDEX = "adjusted rand index"
STATS_OPTIONS = [ADJ_MUTUAL_INFO, FOWLKES_MALLOWS, ADJ_RAND_INDEX]


class ClustererComparisonCalculator(ConfigReader, Statistics):
    """

    Methods to compute matrix of comparisons between different cluster models.

    .. note::
       `Example of exepcted output <https://github.com/sgoldenlab/simba/blob/master/misc/clusterer_comparisons_20240401112101.xlsx>`__.

    :example:
    >>> cluster_comparer = ClustererComparisonCalculator(config_path="/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/project_config.ini", data_dir='/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/clusters', statistics=STATS_OPTIONS)
    >>> cluster_comparer.run()
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        data_dir: Union[str, os.PathLike],
        statistics: List[Literal[ADJ_MUTUAL_INFO, FOWLKES_MALLOWS, ADJ_RAND_INDEX]],
    ):

        ConfigReader.__init__(
            self, config_path=config_path, read_video_info=False, create_logger=False
        )
        check_if_dir_exists(in_dir=data_dir)
        check_file_exist_and_readable(file_path=config_path)
        self.data_paths = find_files_of_filetypes_in_directory(
            directory=data_dir, extensions=[f".{Formats.PICKLE.value}"]
        )
        if len(self.data_paths) < 2:
            raise CountError(
                msg=f"Cluster comparisons require at least two models. Found {len(self.data_paths)} in {data_dir}",
                source=self.__class__.__name__,
            )
        check_valid_lst(
            data=statistics,
            source=self.__class__.__name__,
            valid_dtypes=(str,),
            valid_values=STATS_OPTIONS,
        )
        self.statistics, self.config_path, self.data_dir = (
            statistics,
            config_path,
            data_dir,
        )
        self.save_path = os.path.join(
            self.logs_path,
            f"clusterer_comparisons_{self.datetime}.{Formats.XLXS.value}",
        )
        with pd.ExcelWriter(self.save_path, mode="w") as writer:
            pd.DataFrame().to_excel(writer, sheet_name=" ", index=True)

    def __save_results(self, df: pd.DataFrame, name: str):
        with pd.ExcelWriter(self.save_path, mode="a") as writer:
            df.to_excel(writer, sheet_name=name, index=True)

    def run(self):
        self.data, obs_cnts = {}, {}
        for file_cnt, file_path in enumerate(self.data_paths):
            self.data_ = read_pickle(data_path=file_path, verbose=True)
            check_if_keys_exist_in_dict(
                data=self.data_,
                key=[Unsupervised.METHODS.value, Clustering.CLUSTER_MODEL.value],
                name=file_path,
            )
            mdl_name = self.data_[Clustering.CLUSTER_MODEL.value][
                Unsupervised.HASHED_NAME.value
            ]
            cluster_data = self.data_[Clustering.CLUSTER_MODEL.value][
                Unsupervised.MODEL.value
            ].labels_
            self.data[mdl_name] = cluster_data
            obs_cnts[mdl_name] = cluster_data.shape[0]

        counts = list(set([v for k, v in obs_cnts.items()]))
        if len(counts) > 1:
            raise CountError(
                msg=f"Cluster comparisons require models built using the data. Found different number of observations in the different mdoels in {self.data_dir}. {obs_cnts}",
                source=self.__class__.__name__,
            )

        for statistic in self.statistics:
            results = {}
            for i in self.data.keys():
                print(f"Computing {statistic} statistics for {i}...")
                results[i] = {}
                for j in self.data.keys():
                    x, y = self.data[i], self.data[j]
                    if statistic == ADJ_RAND_INDEX:
                        s = Statistics.adjusted_rand(x=x, y=y)
                    elif statistic == ADJ_MUTUAL_INFO:
                        s = Statistics.adjusted_mutual_info(x=x, y=y)
                    else:
                        s = Statistics.fowlkes_mallows(x=x, y=y)
                    results[i][j] = s
            df = pd.DataFrame.from_dict(results, orient="index")
            self.__save_results(df=df, name=statistic)

        self.timer.stop_timer()
        stdout_success(
            f"Cluster comparisons saved at {self.save_path}",
            elapsed_time=self.timer.elapsed_time_str,
        )


# x = ClustererComparison(config_path="/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/project_config.ini", data_dir='/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/clusters', statistics=STATS_OPTIONS)
# x.run()



import os
from tkinter import *
from typing import Union

from simba.data_processors.directing_other_animals_calculator import \
    DirectingOtherAnimalsAnalyzer
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import CreateLabelFrameWithIcon, SimbaCheckbox
from simba.utils.enums import Formats, Keys, Links
from simba.utils.errors import (AnimalNumberError, InvalidInputError,
                                NoFilesFoundError)
from simba.utils.warnings import SkippingRuleWarning


class AnimalDirectingAnimalPopUp(ConfigReader, PopUpMixin):

    """
    :example:
    >>> test = AnimalDirectingAnimalPopUp(config_path=r"C:\troubleshooting\two_black_animals_14bp\project_folder\project_config.ini")
    """
    def __init__(self, config_path: Union[str, os.PathLike]):
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        if self.animal_cnt < 2:
            raise AnimalNumberError(msg=f"Directionality between animals require at least two animals. The SimBA project is set to use {self.animal_cnt} animal.", source=self.__class__.__name__,)
        if len(self.outlier_corrected_paths) == 0:
            raise NoFilesFoundError(msg=f"No data files found in {self.outlier_corrected_dir}.", source=self.__class__.__name__)
        PopUpMixin.__init__(self, title="ANALYZE DIRECTIONALITY BETWEEN ANIMALS", size=(400, 400), icon='direction')
        self.settings_frm = CreateLabelFrameWithIcon( parent=self.main_frm, header="OUTPUT FORMATS", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.DIRECTING_ANIMALS_PLOTS.value)

        boolean_tables_cb, self.boolean_tables_var = SimbaCheckbox(parent=self.settings_frm, txt="CREATE BOOLEAN TABLES", txt_img='table')
        summary_table_cb, self.summary_tables_var = SimbaCheckbox(parent=self.settings_frm, txt="CREATE DETAILED SUMMARY TABLES (INCLUDING COORDINATES)", txt_img='table')
        aggregate_statistics_cb, self.aggregate_statistics_var = SimbaCheckbox(parent=self.settings_frm, txt="CREATE AGGREGATE STATISTICS TABLE", txt_img='table')
        append_cb, self.append_var = SimbaCheckbox(parent=self.settings_frm, txt="APPEND BOOLEAN TABLES TO FEATURES", txt_img='table')

        self.settings_frm.grid(row=0, column=0, sticky="NW")
        boolean_tables_cb.grid(row=0, column=0, sticky="NW")
        summary_table_cb.grid(row=1, column=0, sticky="NW")
        aggregate_statistics_cb.grid(row=2, column=0, sticky="NW")
        append_cb.grid(row=3, column=0, sticky="NW")
        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()

    def run(self):
        if (not self.boolean_tables_var.get() and not self.summary_tables_var.get() and not self.aggregate_statistics_var.get()):
            raise InvalidInputError("Boolean tables, summary tables, and aggregate statistics options are all UNCHECKED. Please select at least one of these output format.", source=self.__class__.__name__,)
        append = self.append_var.get()
        if not self.boolean_tables_var.get() and self.append_var.get():
            SkippingRuleWarning(msg='To append boolean tables to features, select create boolean tables as well as the checkbox for appending boolean tables to features.')
            append = False

        directing_animals_analyzer = DirectingOtherAnimalsAnalyzer(config_path=self.config_path,
                                                                   bool_tables=self.boolean_tables_var.get(),
                                                                   summary_tables=self.summary_tables_var.get(),
                                                                   aggregate_statistics_tables=self.aggregate_statistics_var.get(),
                                                                   append_bool_tables_to_features=append)
        directing_animals_analyzer.run()


#test = AnimalDirectingAnimalPopUp(config_path=r"C:\troubleshooting\two_black_animals_14bp\project_folder\project_config.ini")



import os
from typing import Optional, Tuple, Union

import cv2
import numpy as np

from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_valid_img, check_if_valid_rgb_tuple,
                                check_int, check_str)
from simba.utils.enums import Options
from simba.utils.errors import InvalidFileTypeError
from simba.utils.read_write import get_fn_ext, get_video_meta_data
from simba.utils.warnings import CropWarning


class ROISelector(object):
    """
    A class for selecting and reflecting Regions of Interest (ROI) in an image.
    The selected region variables are stored in self: top_left, bottom_right, width, height etc.

    .. image:: _static/img/roi_selector.gif
       :width: 600
       :align: center

    :param Union[str, os.PathLike] path: Path to the image or video file. Can also be an image represented as a numpy array.
    :param int thickness: Thickness of the rectangle border for visualizing the ROI.
    :param Tuple[int, int, int] clr: BGR color tuple for visualizing the ROI. Default: deep pink.
    :param Optional[str] title: Title of the drawing window. If None, then `Draw ROI - Press ESC when drawn`.
    :param bool destroy: If True, destroy drawing window when completed. Default: True.

    :example:
    >>> img_selector = ROISelector(path='/Users/simon/Desktop/compute_overlap.png', clr=(0, 255, 0), thickness=2)
    >>> img_selector.run()
    """

    def __init__(self,
                 path: Union[str, os.PathLike, np.ndarray],
                 thickness: int = 10,
                 clr: Tuple[int, int, int] = (147, 20, 255),
                 title: Optional[str] = None,
                 destroy: bool = True) -> None:

        check_if_valid_rgb_tuple(data=clr)
        check_int(name="Thickness", value=thickness, min_value=1, raise_error=True)
        if title is not None:
            check_str(name="ROI title window", value=title)
        self.roi_start = None
        self.roi_end = None
        self.selecting_roi = False
        self.clr = clr
        self.complete = False
        self.terminate = False
        self.thickness = int(thickness)
        self.destroy = destroy

        if isinstance(path, np.ndarray):
            self.image = path
            check_if_valid_img(data=self.image, source=self.__class__.__name__, raise_error=True)
        else:
            check_file_exist_and_readable(file_path=path)
            _, filename, ext = get_fn_ext(filepath=path)
            if ext in Options.ALL_VIDEO_FORMAT_OPTIONS.value:
                _ = get_video_meta_data(video_path=path)
                cap = cv2.VideoCapture(path)
                cap.set(1, 0)
                _, self.image = cap.read()

            elif ext in Options.ALL_IMAGE_FORMAT_OPTIONS.value:
                self.image = cv2.imread(path).astype(np.uint8)

            else:
                raise InvalidFileTypeError(msg=f"Cannot crop a {ext} file.", source=self.__class__.__name__)
            if title is None:
                title = f"Draw ROI video {filename} - Press ESC when drawn"

        self.img_cpy = self.image.copy()
        self.w, self.h = self.image.shape[1], self.image.shape[0]

        if title is None:
            self.title = "Draw ROI - Press ESC when drawn"
        else:
            self.title = title

    def mouse_callback(self, event, x, y, flags, param):
        if event == cv2.EVENT_LBUTTONDOWN:
            self.roi_start = (x, y)
            self.roi_end = (x, y)
            self.selecting_roi = True

        elif event == cv2.EVENT_LBUTTONUP:
            self.roi_end = (x, y)
            self.selecting_roi = False

        elif event == cv2.EVENT_MOUSEMOVE and self.selecting_roi:
            self.roi_end = (x, y)
            self.img_cpy = self.image.copy()
            cv2.rectangle(self.img_cpy, self.roi_start, self.roi_end, self.clr, self.thickness)

    def run(self):
        cv2.namedWindow(self.title, cv2.WINDOW_NORMAL)
        cv2.setMouseCallback(self.title, self.mouse_callback)
        while not self.terminate:
            if self.selecting_roi or self.img_cpy is None:
                self.img_cpy = self.image.copy()
                if self.roi_start is not None and self.roi_end is not None:
                    cv2.rectangle(self.img_cpy, self.roi_start, self.roi_end, self.clr, self.thickness)

            cv2.imshow(self.title, self.img_cpy)
            key = cv2.waitKey(1) & 0xFF

            if (key == ord("c") and self.roi_start is not None and self.roi_end is not None):
                selected_roi = self.image[self.roi_start[1] : self.roi_end[1], self.roi_start[0] : self.roi_end[0]].copy()
                reflected_roi = cv2.flip(selected_roi, 1)
                self.image[self.roi_start[1] : self.roi_end[1], self.roi_start[0] : self.roi_end[0]] = reflected_roi
                self.roi_start = None
                self.roi_end = None

            elif key == 27:
                if self.run_checks():
                    if self.destroy:
                        cv2.destroyAllWindows()
                    cv2.waitKey(1)
                    self.terminate = True
                    break

            if cv2.getWindowProperty(self.title, cv2.WND_PROP_VISIBLE) < 1:
                if self.destroy:
                    cv2.destroyAllWindows()
                    self.terminate = True
                break

        if self.terminate:
            self.run_checks()

    def run_checks(self):
        self.top_left = min(self.roi_start[0], self.roi_end[0]), min(self.roi_start[1], self.roi_end[1])
        self.bottom_right = max(self.roi_start[0], self.roi_end[0]), max(self.roi_start[1], self.roi_end[1])
        if self.top_left[0] < 0:
            self.top_left = (0, self.top_left[1])
        if self.top_left[1] < 0:
            self.top_left = (self.top_left[0], 0)
        if self.bottom_right[0] < 0:
            self.bottom_right = (0, self.bottom_right[1])
        if self.bottom_right[1] < 0:
            self.bottom_right = (self.bottom_right[0], 0)
        if self.bottom_right[0] > self.w:
            self.bottom_right = (self.w, self.bottom_right[1])
        if self.bottom_right[1] > self.h:
            self.bottom_right = (self.bottom_right[0], self.h)
        self.width = self.bottom_right[0] - self.top_left[0]
        self.height = self.bottom_right[1] - self.top_left[1]
        self.center = (int(self.top_left[0] + (self.width / 2)), int(self.top_left[1] + (self.height / 2)))

        self.bottom_right_tag = (int(self.top_left[0] + self.width), int(self.top_left[1] + self.height))
        self.top_right_tag = (int(self.top_left[0] + self.width), int(self.top_left[1]))
        self.bottom_left_tag = (int(self.top_left[0]), int(self.top_left[1] + self.height))
        self.top_tag = (int(self.top_left[0] + self.width / 2), int(self.top_left[1]))
        self.right_tag = (int(self.top_left[0] + self.width), int(self.top_left[1] + self.height / 2))
        self.left_tag = (int(self.top_left[0]), int(self.top_left[1] + self.height / 2))
        self.bottom_tag = (int(self.top_left[0] + self.width / 2), int(self.top_left[1] + self.height))

        self.complete = True

        if (self.width == 0 and self.height == 0) or (self.width + self.height + self.top_left[0] + self.top_left[1] == 0):
            CropWarning(msg="CROP WARNING: ROI height and width are both 0. Please try again.", source=self.__class__.__name__)
            return False
        else:
            return True


# img_selector = ROISelector(path=r"D:\weights\2025-04-17_17-17-14\cropped_slowed_output.mp4", clr=(0, 255, 0), thickness=2, destroy=False)
# img_selector.run()
# print(img_selector.top_right_tag, img_selector.width, img_selector.height)


__author__ = "Simon Nilsson"

import os
from copy import deepcopy
from typing import Optional, Tuple, Union

import cv2
import numpy as np
import pandas as pd

from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists, check_if_valid_rgb_tuple,
                                check_int, check_valid_tuple)
from simba.utils.errors import (DataHeaderError, DuplicationError,
                                InvalidFileTypeError, InvalidInputError)
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (get_fn_ext,
                                    get_number_of_header_columns_in_df,
                                    get_video_meta_data, read_config_file,
                                    read_df)

H5 = ".h5"
CSV = ".csv"


class EzPathPlot(object):
    """
    Create a simpler path plot for a single path in a single video.

    .. note::
       For more refined / complex path plots with/without multiprocessing for inproved speed, see ``simba.plotting.path_plotter.PathPlotterSingleCore`` and ``simba.plotting.path_plotter_mp.PathPlotterMulticore``.

    .. image:: _static/img/EzPathPlot.gif
       :width: 500
       :align: center

    .. image:: _static/img/EzPathPlot_2.gif
       :width: 500
       :align: center

    :param Union[str, os.PathLike] data_path: The path to the data file in H5c or CSV format containing the pose estimation coordinates.
    :param Optional[Union[str, os.PathLike]] video_path: The path to the video file. Optional. If provided, the FPS and size is grabbed from the metadata of the video file. If None, then pass ``fps`` and ``size``.
    :param Optional[Tuple[int, int]] size: Size of the path plot (width, height). Used if video_path is None.
    :param Optional[int] fps: The FPS of the path plot. Used if video_path is None.
    :param str body_part: The specific body part to plot the path for.
    :param Optional[bool] last_frm_only: If True, creates just a single .png image representing the full path in last image in the video.
    :param Optional[Tuple[int, int, int]] bg_color: The background color of the plot. Defaults to (255, 255, 255).
    :param Optional[Tuple[int, int, int]] line_color: The color of the path line. Defaults to (147, 20, 255).
    :param Optional[int] line_thickness: The thickness of the path line. Defaults to 10.
    :param Optional[int] circle_size: The size of the circle indicating each data point. Defaults to 5.
    :param Optional[Union[str, os.PathLike]] save_path: The location to store the path plot. If None, then use the same path as the data path with ``_line_plot`` suffix.


    :example I:
    >>> path_plotter = EzPathPlot(data_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/h5/Together_1DLC_resnet50_two_black_mice_DLC_052820May27shuffle1_150000_el.h5', video_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/videos/Together_1.avi', body_part='Mouse_1_Nose', bg_color=(255, 255, 255), line_color=(147,20,255))
    >>> path_plotter.run()

    :example II:
    >>> path_plotter = EzPathPlot(data_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/h5/Together_1DLC_resnet50_two_black_mice_DLC_052820May27shuffle1_150000_el.h5', size=(2056, 1549), fps=30, body_part='Mouse_1_Nose', bg_color=(255, 255, 255), line_color=(147,20,255))
    >>> path_plotter.run()
    """

    def __init__(
        self,
        data_path: Union[str, os.PathLike],
        body_part: str,
        bg_color: Optional[Tuple[int, int, int]] = (255, 255, 255),
        line_color: Optional[Tuple[int, int, int]] = (147, 20, 255),
        video_path: Optional[Union[str, os.PathLike]] = None,
        size: Optional[Tuple[int, int]] = None,
        fps: Optional[int] = None,
        line_thickness: Optional[int] = 10,
        circle_size: Optional[int] = 5,
        last_frm_only: Optional[bool] = False,
        save_path: Optional[Union[str, os.PathLike]] = None,
    ):

        check_file_exist_and_readable(file_path=data_path)
        check_if_valid_rgb_tuple(data=bg_color)
        check_if_valid_rgb_tuple(data=line_color)
        check_int(
            name=f"{self.__class__.__name__} line_thickness",
            value=line_thickness,
            min_value=1,
        )
        check_int(
            name=f"{self.__class__.__name__} circle_size",
            value=circle_size,
            min_value=1,
        )
        if save_path is not None:
            check_if_dir_exists(
                in_dir=os.path.dirname(save_path), create_if_not_exist=True
            )
        if line_color == bg_color:
            raise DuplicationError(
                msg=f"The line and background cannot be identical - ({line_color})",
                source=self.__class__.__name__,
            )
        if video_path is not None:
            video_meta_data = get_video_meta_data(video_path=video_path)
            self.height, self.width = int(video_meta_data["height"]), int(
                video_meta_data["width"]
            )
            self.fps = int(video_meta_data["fps"])
        else:
            if (size is None) or (fps is None):
                raise InvalidInputError(
                    msg="If video path is None, then pass size and fps",
                    source=self.__class__.__name__,
                )
            check_valid_tuple(
                x=size,
                source=f"{self.__class__.__name__} size",
                accepted_lengths=(2,),
                valid_dtypes=(int,),
            )
            self.height, self.width = size[1], size[0]
            check_int(name=f"{self.__class__.__name__} fps", value=fps, min_value=1)
            self.fps = int(fps)
        dir, file_name, ext = get_fn_ext(filepath=data_path)
        if ext.lower() == H5:
            self.data = pd.read_hdf(data_path)
            headers = []
            if len(self.data.columns[0]) == 4:
                for c in self.data.columns:
                    headers.append("{}_{}_{}".format(c[1], c[2], c[3]))
            elif len(self.data.columns[0]) == 3:
                for c in self.data.columns:
                    headers.append("{}_{}".format(c[2], c[3]))
            self.data.columns = headers
        elif ext.lower() == CSV:
            self.data = pd.read_csv(data_path)
        else:
            raise InvalidFileTypeError(
                msg=f"File type {ext} is not supported (OPTIONS: h5 or csv)"
            )
        if len(self.data.columns[0]) == 4:
            self.data = self.data.loc[3:]
        elif len(self.data.columns[0]) == 3:
            self.data = self.data.loc[2:]
        body_parts_available = list(set([x[:-2] for x in self.data.columns]))
        if body_part not in body_parts_available:
            raise DataHeaderError(
                msg=f"Body-part {body_part} is not present in the data file. The body-parts available are: {body_parts_available}",
                source=self.__class__.__name__,
            )
        bps = [f"{body_part}_x", f"{body_part}_y"]
        if (bps[0] not in self.data.columns) or (bps[1] not in self.data.columns):
            raise DataHeaderError(
                msg=f"Could not finc column {bps[0]} and/or column {bps[1]} in the data file {data_path}",
                source=self.__class__.__name__,
            )
        self.data = (
            self.data[bps]
            .fillna(method="ffill")
            .astype(int)
            .reset_index(drop=True)
            .values
        )
        if (save_path is None) and (not last_frm_only):
            self.save_name = os.path.join(dir, f"{file_name}_line_plot.mp4")
        elif (save_path is None) and (last_frm_only):
            self.save_name = os.path.join(dir, f"{file_name}_line_plot.png")
        else:
            self.save_name = save_path
        self.bg_img = np.zeros([self.height, self.width, 3])
        self.bg_img[:] = [bg_color]
        self.line_color, self.line_thickness, self.circle_size, self.last_frm = (
            line_color,
            line_thickness,
            circle_size,
            last_frm_only,
        )
        self.timer = SimbaTimer(start=True)

    def run(self):
        if not self.last_frm:
            self.writer = cv2.VideoWriter(
                self.save_name, 0x7634706D, self.fps, (self.width, self.height)
            )
            for i in range(1, self.data.shape[0] + 1):
                line_data = self.data[: i + 1]
                img = deepcopy(self.bg_img)
                for j in range(1, line_data.shape[0]):
                    x1, y1 = line_data[j - 1][0], line_data[j - 1][1]
                    x2, y2 = line_data[j][0], line_data[j][1]
                    cv2.line(
                        img, (x1, y1), (x2, y2), self.line_color, self.line_thickness
                    )
                cv2.circle(
                    img,
                    (line_data[-1][0], line_data[-1][1]),
                    self.circle_size,
                    self.line_color,
                    -1,
                )
                self.writer.write(img.astype(np.uint8))
                print(f"Frame {i}/{len(self.data)} complete...")
            self.writer.release()
        else:
            img = deepcopy(self.bg_img)
            for j in range(1, self.data.shape[0]):
                x1, y1 = self.data[j - 1][0], self.data[j - 1][1]
                x2, y2 = self.data[j][0], self.data[j][1]
                cv2.line(img, (x1, y1), (x2, y2), self.line_color, self.line_thickness)
            cv2.imwrite(filename=self.save_name, img=img)
        self.timer.stop_timer()
        stdout_success(
            msg=f"Path plot saved at {self.save_name}",
            elapsed_time=self.timer.elapsed_time_str,
        )


# path_plotter = EzPathPlot(data_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/h5/Together_1DLC_resnet50_two_black_mice_DLC_052820May27shuffle1_150000_el.h5',
#                           size=(2056, 1549),
#                           fps=30,
#                           body_part='Mouse_1_Nose',
#                           bg_color=(255, 255, 255),
#                           last_frm_only=False,
#                           line_color=(147,20,255))
# path_plotter.run()



__author__ = "Simon Nilsson"

import threading
from tkinter import *

from simba.mixins.pop_up_mixin import PopUpMixin
from simba.third_party_label_appenders.converters import labelme_to_img_dir
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, FolderSelect,
                                        SimBADropDown)
from simba.utils.read_write import str_2_bool

IMG_FORMATS = ['png', 'jpeg', 'bmp', 'webp']

class Labelme2ImgsPopUp(PopUpMixin):

    """
    :example:
    >>> Labelme2ImgsPopUp()
    """
    def __init__(self):
        super().__init__(title="LABELME TO IMAGES")
        settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SETTINGS", icon_name='settings')
        self.labelme_dir = FolderSelect(settings_frm, folderDescription="LABELME DIRECTORY:", lblwidth=35, width=40)
        self.save_dir = FolderSelect(settings_frm, folderDescription="SAVE DIRECTORY:", lblwidth=35, width=40)
        self.img_format_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=IMG_FORMATS, label="IMAGE FORMAT: ", label_width=35, dropdown_width=40, value=IMG_FORMATS[0])
        self.grey_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="GREYSCALE: ", label_width=35, dropdown_width=40, value='FALSE')


        settings_frm.grid(row=0, column=0, sticky=NW)
        self.labelme_dir.grid(row=0, column=0, sticky=NW)
        self.save_dir.grid(row=1, column=0, sticky=NW)
        self.img_format_dropdown.grid(row=2, column=0, sticky=NW)
        self.grey_dropdown.grid(row=3, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()

    def run(self):
        labelme_dir = self.labelme_dir.folder_path
        save_dir = self.save_dir.folder_path
        img_format = str(self.img_format_dropdown.get_value())
        grey = str_2_bool(self.grey_dropdown.get_value())
        thread = threading.Thread(target=labelme_to_img_dir, kwargs={'labelme_dir': labelme_dir, 'img_dir': save_dir, 'img_format': img_format, 'greyscale': grey})
        thread.start()



__author__ = "Simon Nilsson"

import ast
import os
from typing import Optional, Union

import pandas as pd
from sklearn.model_selection import train_test_split

from simba.mixins.config_reader import ConfigReader
from simba.mixins.train_model_mixin import TrainModelMixin
from simba.utils.checks import check_if_filepath_list_is_empty, check_int
from simba.utils.enums import (ConfigKey, Dtypes, Formats, Methods,
                               MLParamKeys, Options)
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import read_config_entry, str_2_bool, write_df


class TrainRandomForestClassifier(ConfigReader, TrainModelMixin):
    """
    Train a single random forest model using hyperparameter setting and evaluation methods
    stored within the SimBA project config .ini file (``global environment``).

    :param Union[str, os.PathLike] config_path: path to SimBA project config file in Configparser format

    .. note::
       `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario1.md#step-7-train-machine-model>`_

    :example:
    >>> model_trainer = TrainRandomForestClassifier(config_path='MyConfigPath')
    >>> model_trainer.run()
    >>> model_trainer.save()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike]):

        ConfigReader.__init__(self, config_path=config_path, create_logger=False)
        TrainModelMixin.__init__(self)
        self.read_model_settings_from_config(config=self.config)
        check_if_filepath_list_is_empty(filepaths=self.target_file_paths, error_msg=f"Zero annotation files found in directory {self.targets_folder}, cannot create model.")
        self.bp_config = read_config_entry(config=self.config, section=ConfigKey.CREATE_ENSEMBLE_SETTINGS.value, option=ConfigKey.POSE_SETTING.value, default_value='user_defined', data_type=Dtypes.STR.value)
        print(f"Reading in {len(self.target_file_paths)} annotated files...")
        self.data_df, self.frm_idx = self.read_all_files_in_folder_mp_futures(self.target_file_paths, self.file_type, [self.clf_name])
        self.frm_idx = pd.DataFrame({"VIDEO": list(self.data_df.index), "FRAME_IDX": self.frm_idx})
        self.data_df = self.check_raw_dataset_integrity(df=self.data_df, logs_path=self.logs_path)
        self.data_df_wo_cords = self.drop_bp_cords(df=self.data_df)
        annotation_cols_to_remove = self.read_in_all_model_names_to_remove(self.config, self.clf_cnt, self.clf_name)
        self.x_y_df = self.delete_other_annotation_columns(self.data_df_wo_cords, list(annotation_cols_to_remove))
        self.class_names = ["Not_" + self.clf_name, self.clf_name]
        self.x_df, self.y_df = self.split_df_to_x_y(self.x_y_df, self.clf_name)
        self.feature_names = self.x_df.columns
        self.check_sampled_dataset_integrity(x_df=self.x_df, y_df=self.y_df)
        print(f"Number of features in dataset: {len(self.x_df.columns)}")
        print(f"Number of {self.clf_name} frames in dataset: {int(self.y_df.sum())} ({str(round(self.y_df.sum() / len(self.y_df), 4) * 100)}%)")

    def perform_sampling(self):
        """
        Method for sampling data for training and testing, and perform over and under-sampling of the training sets
        as indicated within the SimBA project config.
        """

        if self.split_type == Methods.SPLIT_TYPE_FRAMES.value:
            self.x_train, self.x_test, self.y_train, self.y_test = train_test_split(self.x_df, self.y_df, test_size=self.tt_size)
        elif self.split_type == Methods.SPLIT_TYPE_BOUTS.value:
            self.x_train, self.x_test, self.y_train, self.y_test = ( self.bout_train_test_splitter(x_df=self.x_df, y_df=self.y_df, test_size=self.tt_size))
        if self.under_sample_setting == Methods.RANDOM_UNDERSAMPLE.value.lower():
            self.x_train, self.y_train = self.random_undersampler(self.x_train, self.y_train, float(self.under_sample_ratio))
        if self.over_sample_setting == Methods.SMOTEENN.value.lower():
            self.x_train, self.y_train = self.smoteen_oversampler(self.x_train, self.y_train, float(self.over_sample_ratio))
        elif self.over_sample_setting == Methods.SMOTE.value.lower():
            self.x_train, self.y_train = self.smote_oversampler(self.x_train, self.y_train, float(self.over_sample_ratio))

        if self.save_train_test_frm_info:
            train_data = self.frm_idx[self.frm_idx.index.isin(self.x_train.index)].set_index("VIDEO")
            test_data = self.frm_idx[self.frm_idx.index.isin(self.x_test.index)].set_index("VIDEO")
            write_df(df=train_data, file_type=Formats.CSV.value, save_path=os.path.join(self.eval_out_path, "train_idx.csv"))
            write_df(df=test_data, file_type=Formats.CSV.value, save_path=os.path.join(self.eval_out_path, "test_idx.csv"))
            print(f"Frame index for the train and test sets saved in {self.eval_out_path} directory...")

    def run(self):
        """
        Method for training single random forest model.
        """
        print("Training and evaluating model...")
        self.timer = SimbaTimer(start=True)
        self.perform_sampling()
        if self.algo == "RF":
            n_estimators = read_config_entry(self.config,ConfigKey.CREATE_ENSEMBLE_SETTINGS.value,MLParamKeys.RF_ESTIMATORS.value,data_type=Dtypes.INT.value)
            max_features = read_config_entry(self.config,ConfigKey.CREATE_ENSEMBLE_SETTINGS.value,MLParamKeys.RF_MAX_FEATURES.value,data_type=Dtypes.STR.value)
            if max_features == "None": max_features = None
            criterion = read_config_entry(self.config, ConfigKey.CREATE_ENSEMBLE_SETTINGS.value, MLParamKeys.RF_CRITERION.value, data_type=Dtypes.STR.value, options=Options.CLF_CRITERION.value)
            min_sample_leaf = read_config_entry(self.config, ConfigKey.CREATE_ENSEMBLE_SETTINGS.value, MLParamKeys.MIN_LEAF.value, data_type=Dtypes.INT.value)
            compute_permutation_importance = read_config_entry(self.config, ConfigKey.CREATE_ENSEMBLE_SETTINGS.value, MLParamKeys.PERMUTATION_IMPORTANCE.value, data_type=Dtypes.STR.value, default_value=False)
            generate_learning_curve = read_config_entry(self.config, ConfigKey.CREATE_ENSEMBLE_SETTINGS.value, MLParamKeys.LEARNING_CURVE.value, data_type=Dtypes.STR.value, default_value=False)
            generate_precision_recall_curve = read_config_entry(self.config, ConfigKey.CREATE_ENSEMBLE_SETTINGS.value, MLParamKeys.PRECISION_RECALL.value, data_type=Dtypes.STR.value, default_value=False)
            generate_example_decision_tree = read_config_entry(self.config, ConfigKey.CREATE_ENSEMBLE_SETTINGS.value, MLParamKeys.EX_DECISION_TREE.value, data_type=Dtypes.STR.value, default_value=False)
            generate_classification_report = read_config_entry(self.config, ConfigKey.CREATE_ENSEMBLE_SETTINGS.value, MLParamKeys.CLF_REPORT.value, data_type=Dtypes.STR.value, default_value=False)
            generate_features_importance_log = read_config_entry(self.config, ConfigKey.CREATE_ENSEMBLE_SETTINGS.value, MLParamKeys.IMPORTANCE_LOG.value, data_type=Dtypes.STR.value, default_value=False)
            generate_features_importance_bar_graph = read_config_entry(self.config,ConfigKey.CREATE_ENSEMBLE_SETTINGS.value,MLParamKeys.IMPORTANCE_LOG.value,data_type=Dtypes.STR.value,default_value=False)
            generate_example_decision_tree_fancy = read_config_entry(self.config,ConfigKey.CREATE_ENSEMBLE_SETTINGS.value,MLParamKeys.EX_DECISION_TREE_FANCY.value,data_type=Dtypes.STR.value,default_value=False)
            generate_shap_scores = read_config_entry(self.config,ConfigKey.CREATE_ENSEMBLE_SETTINGS.value,MLParamKeys.SHAP_SCORES.value,data_type=Dtypes.STR.value,default_value=False)
            save_meta_data = read_config_entry(self.config,ConfigKey.CREATE_ENSEMBLE_SETTINGS.value,MLParamKeys.RF_METADATA.value,data_type=Dtypes.STR.value,default_value=False)
            compute_partial_dependency = read_config_entry(self.config,ConfigKey.CREATE_ENSEMBLE_SETTINGS.value,MLParamKeys.PARTIAL_DEPENDENCY.value,data_type=Dtypes.STR.value, default_value=False)
            cuda = str_2_bool(read_config_entry(self.config, ConfigKey.CREATE_ENSEMBLE_SETTINGS.value, MLParamKeys.CUDA.value, data_type=Dtypes.STR.value, default_value=False))

            if self.config.has_option(ConfigKey.CREATE_ENSEMBLE_SETTINGS.value, MLParamKeys.CLASS_WEIGHTS.value):
                class_weights = read_config_entry(self.config, ConfigKey.CREATE_ENSEMBLE_SETTINGS.value, MLParamKeys.CLASS_WEIGHTS.value, data_type=Dtypes.STR.value, default_value=Dtypes.NONE.value)
                if class_weights == "custom":
                    class_weights = ast.literal_eval(read_config_entry(self.config,ConfigKey.CREATE_ENSEMBLE_SETTINGS.value,MLParamKeys.CLASS_CUSTOM_WEIGHTS.value,data_type=Dtypes.STR.value))
                    for k, v in class_weights.items():
                        class_weights[k] = int(v)
                if class_weights == Dtypes.NONE.value:
                    class_weights = None
            else:
                class_weights = None

            if generate_learning_curve in Options.PERFORM_FLAGS.value:
                shuffle_splits = read_config_entry(self.config,ConfigKey.CREATE_ENSEMBLE_SETTINGS.value,MLParamKeys.LEARNING_CURVE_K_SPLITS.value,data_type=Dtypes.INT.value,default_value=Dtypes.NAN.value)
                dataset_splits = read_config_entry(self.config,ConfigKey.CREATE_ENSEMBLE_SETTINGS.value,MLParamKeys.LEARNING_DATA_SPLITS.value,data_type=Dtypes.INT.value,default_value=Dtypes.NAN.value)
                check_int(name=MLParamKeys.LEARNING_CURVE_K_SPLITS.value, value=shuffle_splits)
                check_int(name=MLParamKeys.LEARNING_DATA_SPLITS.value, value=dataset_splits)
            else:
                shuffle_splits, dataset_splits = Dtypes.NAN.value, Dtypes.NAN.value
            if generate_features_importance_bar_graph in Options.PERFORM_FLAGS.value:
                feature_importance_bars = read_config_entry(self.config, ConfigKey.CREATE_ENSEMBLE_SETTINGS.value, MLParamKeys.IMPORTANCE_BARS_N.value, Dtypes.INT.value, Dtypes.NAN.value)
                check_int(name=MLParamKeys.IMPORTANCE_BARS_N.value, value=feature_importance_bars, min_value=1)
            else:
                feature_importance_bars = Dtypes.NAN.value
            (shap_target_present_cnt, shap_target_absent_cnt, shap_save_n, shap_multiprocess) = (None, None, None, None)
            if generate_shap_scores in Options.PERFORM_FLAGS.value:
                shap_target_present_cnt = read_config_entry(self.config, ConfigKey.CREATE_ENSEMBLE_SETTINGS.value, MLParamKeys.SHAP_PRESENT.value, data_type=Dtypes.INT.value, default_value=0)
                shap_target_absent_cnt = read_config_entry(self.config, ConfigKey.CREATE_ENSEMBLE_SETTINGS.value, MLParamKeys.SHAP_ABSENT.value, data_type=Dtypes.INT.value, default_value=0)
                shap_save_n = read_config_entry(self.config, ConfigKey.CREATE_ENSEMBLE_SETTINGS.value, MLParamKeys.SHAP_SAVE_ITERATION.value, data_type=Dtypes.STR.value, default_value=Dtypes.NONE.value)
                shap_multiprocess = read_config_entry(self.config, ConfigKey.CREATE_ENSEMBLE_SETTINGS.value, MLParamKeys.SHAP_MULTIPROCESS.value, data_type=Dtypes.STR.value, default_value="False")
                try:
                    shap_save_n = int(shap_save_n)
                except ValueError:
                    shap_save_n = shap_target_present_cnt + shap_target_absent_cnt
                check_int(name=MLParamKeys.SHAP_PRESENT.value, value=shap_target_present_cnt)
                check_int(name=MLParamKeys.SHAP_ABSENT.value, value=shap_target_absent_cnt)



            self.rf_clf = self.clf_define(n_estimators=n_estimators,
                                          max_depth=self.rf_max_depth,
                                          max_features=max_features,
                                          n_jobs=-1,
                                          criterion=criterion,
                                          min_samples_leaf=min_sample_leaf,
                                          verbose=1,
                                          class_weight=class_weights,
                                          cuda=cuda)

            print(f"Fitting {self.clf_name} model...")
            self.rf_clf = self.clf_fit(clf=self.rf_clf, x_df=self.x_train, y_df=self.y_train)

            if compute_permutation_importance in Options.PERFORM_FLAGS.value:
                self.calc_permutation_importance(x_test=self.x_test, y_test=self.y_test, clf=self.rf_clf, feature_names=self.feature_names, clf_name=self.clf_name, save_dir=self.eval_out_path, save_file_no=None, plot=True)

            if generate_learning_curve in Options.PERFORM_FLAGS.value:
                self.calc_learning_curve(x_y_df=self.x_y_df, clf_name=self.clf_name, shuffle_splits=shuffle_splits, dataset_splits=dataset_splits, tt_size=self.tt_size, rf_clf=self.rf_clf, save_dir=self.eval_out_path)

            if generate_precision_recall_curve in Options.PERFORM_FLAGS.value:
                self.calc_pr_curve(self.rf_clf,self.x_test,self.y_test,self.clf_name,self.eval_out_path)
            if generate_example_decision_tree in Options.PERFORM_FLAGS.value:
                self.create_example_dt(self.rf_clf, self.clf_name, self.feature_names, self.class_names, self.eval_out_path)
            if generate_classification_report in Options.PERFORM_FLAGS.value:
                self.create_clf_report(self.rf_clf, self.x_test, self.y_test, self.class_names, self.eval_out_path)
            if generate_features_importance_log in Options.PERFORM_FLAGS.value:
                self.create_x_importance_log(self.rf_clf, self.feature_names, self.clf_name, self.eval_out_path)
            if generate_features_importance_bar_graph in Options.PERFORM_FLAGS.value:
                self.create_x_importance_bar_chart(self.rf_clf, self.feature_names, self.clf_name, self.eval_out_path, feature_importance_bars)
            if generate_example_decision_tree_fancy in Options.PERFORM_FLAGS.value:
                self.dviz_classification_visualization(self.x_train, self.y_train, self.clf_name, self.class_names, self.eval_out_path)
            if generate_shap_scores in Options.PERFORM_FLAGS.value:
                shap_plot = self.bp_config in {'14', '16'}
                if not shap_multiprocess in Options.PERFORM_FLAGS.value:
                    self.create_shap_log(rf_clf=self.rf_clf,
                                         x=self.x_train,
                                         y=self.y_train,
                                         x_names=list(self.feature_names),
                                         clf_name=self.clf_name,
                                         cnt_present=shap_target_present_cnt,
                                         cnt_absent=shap_target_absent_cnt,
                                         verbose=True,
                                         plot=shap_plot,
                                         save_it=shap_save_n,
                                         save_dir=self.eval_out_path)
                else:
                    self.create_shap_log_concurrent_mp(rf_clf=self.rf_clf,
                                                       x=self.x_train,
                                                       y=self.y_train,
                                                       x_names=list(self.feature_names),
                                                       clf_name=self.clf_name,
                                                       cnt_present=shap_target_present_cnt,
                                                       cnt_absent=shap_target_absent_cnt,
                                                       save_dir=self.eval_out_path,
                                                       plot=shap_plot)

            if compute_partial_dependency in Options.PERFORM_FLAGS.value:
                self.partial_dependence_calculator(clf=self.rf_clf, x_df=self.x_train, clf_name=self.clf_name, save_dir=self.eval_out_path)

            if save_meta_data in Options.PERFORM_FLAGS.value:
                print("Saving model meta data file...")
                save_path = os.path.join(self.eval_out_path, f"{self.clf_name}_meta.csv")
                table = {MLParamKeys.CLASSIFIER_NAME.value:                         self.clf_name,
                         MLParamKeys.RF_CRITERION.value:                            criterion,
                         MLParamKeys.RF_MAX_FEATURES.value:                         max_features,
                         MLParamKeys.MIN_LEAF.value:                                min_sample_leaf,
                         MLParamKeys.RF_ESTIMATORS.value:                           n_estimators,
                         MLParamKeys.PERMUTATION_IMPORTANCE.value:                  compute_permutation_importance,
                         MLParamKeys.CLF_REPORT.value:                              generate_classification_report,
                         MLParamKeys.EX_DECISION_TREE.value:                        generate_example_decision_tree,
                         MLParamKeys.IMPORTANCE_BAR_CHART.value:                    generate_features_importance_bar_graph,
                         MLParamKeys.IMPORTANCE_LOG.value:                          generate_features_importance_log,
                         MLParamKeys.PRECISION_RECALL.value:                        generate_precision_recall_curve,
                         MLParamKeys.RF_METADATA.value:                             save_meta_data,
                         MLParamKeys.LEARNING_CURVE.value:                          generate_learning_curve,
                         MLParamKeys.LEARNING_CURVE_DATA_SPLITS.value:              dataset_splits,
                         MLParamKeys.LEARNING_CURVE_K_SPLITS.value:                 shuffle_splits,
                         MLParamKeys.N_FEATURE_IMPORTANCE_BARS.value:               feature_importance_bars,
                         MLParamKeys.OVERSAMPLE_RATIO.value:                        self.over_sample_ratio,
                         MLParamKeys.OVERSAMPLE_SETTING.value:                      self.over_sample_setting,
                         MLParamKeys.TT_SIZE.value:                                 self.tt_size,
                         MLParamKeys.TRAIN_TEST_SPLIT_TYPE.value:                   self.split_type,
                         MLParamKeys.UNDERSAMPLE_RATIO.value:                       self.under_sample_ratio,
                         MLParamKeys.UNDERSAMPLE_SETTING.value:                     self.under_sample_setting,
                         MLParamKeys.CLASS_WEIGHTS.value:                           self.rf_max_depth,
                         MLParamKeys.CUDA.value:                                    cuda}

                out_df = pd.DataFrame([list(table.values())], columns=list(table.keys()))
                out_df.to_csv(save_path)

    def save(self) -> None:
        """
        Method for saving pickled RF model. The model is saved in the `models/generated_models` directory
        of the SimBA project tree.
        """

        self.timer.stop_timer()
        if not os.listdir(self.model_dir_out):
            os.makedirs(self.model_dir_out)
        self.save_rf_model(self.rf_clf, self.clf_name, self.model_dir_out)
        stdout_success(msg=f"Classifier {self.clf_name} saved in {self.model_dir_out} directory", elapsed_time=self.timer.elapsed_time_str, source=self.__class__.__name__)
        stdout_success(msg=f"Evaluation files are in {self.eval_out_path} folders", source=self.__class__.__name__)



# test = TrainRandomForestClassifier(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini")
# test.run()
# test.save()

# test = TrainRandomForestClassifier(config_path=r"/mnt/c/troubleshooting/mitra/project_folder/project_config.ini")
# test.run()
# test.save()

# test = TrainRandomForestClassifier(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini")
# test.run()
# test.save()



# test = TrainRandomForestClassifier(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini")
# test.run()
# test.save()



# test = TrainRandomForestClassifier(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini")
# test.run()
# test.save()



#
# test = TrainRandomForestClassifier(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini")
# test.run()
# test.save()


# test = TrainRandomForestClassifier(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/mouse_open_field/project_folder/project_config.ini')
# test.run()
# test.save()


# test = TrainRandomForestClassifier(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# test.perform_sampling()
# test.train_model()
# test.save_model()

# test = TrainSingleModel(config_path='/Users/simon/Desktop/envs/troubleshooting/locomotion/project_folder/project_config.ini')
# test.perform_sampling()
# test.train_model()
# test.save_model()


# test = TrainSingleModel(config_path='/Users/simon/Desktop/envs/troubleshooting/prueba/project_folder/project_config.ini')
# test.perform_sampling()
# test.train_model()
# test.save_model()

# test = TrainSingleModel(config_path='/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/project_config.ini')
# test.perform_sampling()
# test.train_model()
# test.save_model()

# test = TrainSingleModel(config_path='/Users/simon/Desktop/envs/troubleshooting/Lucas/project_folder/project_config.ini')
# test.perform_sampling()
# test.train_model()
# test.save`_model()



import os
from typing import Tuple, Union

import cv2
import numpy as np

from simba.utils.checks import check_if_valid_img, check_instance
from simba.utils.read_write import get_video_meta_data, read_frm_of_video
from simba.utils.warnings import InValidUserInputWarning


class BrightnessContrastUI:
    """
     Create a user interface using OpenCV to explore and change the brightness and contrast of a video.

     .. note::
        Adapted from `geeksforgeeks <https://www.geeksforgeeks.org/changing-the-contrast-and-brightness-of-an-image-using-python-opencv/>`_.

     .. image:: _static/img/brightness_contrast_ui.gif
        :width: 700
        :align: center

     :param Union[str, os.PathLike] video_path: Path to the video file or an image in numpy array format.
     :return Tuple: The scaled brightness and scaled contrast values on scale -1 to +1 and 0-2 respectively, suitable for FFmpeg conversion

     :example I:
     >>> brightness_contrast_ui(video_path='/Users/simon/Desktop/envs/simba/troubleshooting/RAT_NOR/project_folder/frames/output/ROI_features/2022-06-20_NOB_DOT_4.mp4')

     :example II:
     >>> img = cv2.imread('/Users/simon/Downloads/PXL_20240429_222923838.jpg')
     >>> brightness_contrast_ui(data=img)

     """

    def __init__(self, data: Union[str, os.PathLike, np.ndarray]):
        self.WINDOW_NAME = "CONTRAST / BRIGHTNESS: HIT ESC TO CONTINUE"
        self.BRIGHTNESS = 'BRIGHTNESS'
        self.CONTRAST = 'CONTRAST'
        self.last_values = {self.BRIGHTNESS: 255, self.CONTRAST: 127}
        self._trackbars_created = False

        check_instance(source=self.__class__.__name__, instance=data, accepted_types=(np.ndarray, str))
        if isinstance(data, str):
            _ = get_video_meta_data(video_path=data)
            self.original_img = read_frm_of_video(video_path=data, frame_index=0)
        else:
            check_if_valid_img(data=data, source=self.__class__.__name__)
            self.original_img = data

    def _on_trackbar_change(self, _):
        if not self._trackbars_created:
            return

        if cv2.getWindowProperty(self.WINDOW_NAME, cv2.WND_PROP_VISIBLE) < 1:
            return

        try:
            brightness_raw = cv2.getTrackbarPos(self.BRIGHTNESS, self.WINDOW_NAME)
            contrast_raw = cv2.getTrackbarPos(self.CONTRAST, self.WINDOW_NAME)
        except cv2.error:
            return

        if (brightness_raw == self.last_values[self.BRIGHTNESS] and
                contrast_raw == self.last_values[self.CONTRAST]):
            return

        self.last_values[self.BRIGHTNESS] = brightness_raw
        self.last_values[self.CONTRAST] = contrast_raw

        # Convert raw trackbar positions to ffmpeg-like brightness and contrast
        brightness = (brightness_raw - 255) / 255.0  # Scale to [-1, 1]
        contrast = contrast_raw / 127.0  # Scale to [0, 2]

        # Apply eq filter-like brightness/contrast: new_img = (img - 128)*contrast + 128 + brightness*255
        img_float = self.original_img.astype(np.float32)
        buf = (img_float - 128) * contrast + 128 + brightness * 255
        buf = np.clip(buf, 0, 255).astype(np.uint8)

        cv2.imshow(self.WINDOW_NAME, buf)

    def _create_window_and_trackbars(self):
        cv2.namedWindow(self.WINDOW_NAME, cv2.WINDOW_NORMAL)
        cv2.imshow(self.WINDOW_NAME, self.original_img)
        cv2.createTrackbar(self.BRIGHTNESS, self.WINDOW_NAME, self.last_values[self.BRIGHTNESS], 510, self._on_trackbar_change)
        cv2.createTrackbar(self.CONTRAST, self.WINDOW_NAME, self.last_values[self.CONTRAST], 254, self._on_trackbar_change)
        self._trackbars_created = True

    def run(self) -> Tuple[float, float]:
        self._trackbars_created = False
        self._create_window_and_trackbars()

        while True:
            if cv2.getWindowProperty(self.WINDOW_NAME, cv2.WND_PROP_VISIBLE) < 1:
                cv2.destroyAllWindows()
                raise InValidUserInputWarning("Window was closed before selecting values.", source=self.__class__.__name__)

            k = cv2.waitKey(10) & 0xFF
            if k == 27:
                try:
                    brightness_raw = cv2.getTrackbarPos(self.BRIGHTNESS, self.WINDOW_NAME)
                    contrast_raw = cv2.getTrackbarPos(self.CONTRAST, self.WINDOW_NAME)
                except cv2.error:
                    continue

                cv2.destroyAllWindows()

                scaled_brightness = (brightness_raw - 255) / 255.0 # [-1, 1]
                scaled_contrast = contrast_raw / 127.0              # [0, 2]

                if scaled_brightness == 0.0 and scaled_contrast == 1.0:
                    InValidUserInputWarning("Both the selected brightness and contrast are default. Select different values.", source=self.__class__.__name__)
                else:
                    return scaled_brightness, scaled_contrast

#ui = BrightnessContrastUI(data=r"/mnt/c/Users/sroni/OneDrive/Desktop/light_dark_box_eq_20250603150339.mp4")
#ui.run()
# ui = BrightnessContrastUI(data=r"D:\brightness_contrast\F1_2.mp4")
# ui.run()